from __future__ import annotations

import contextlib
import dataclasses
import enum
import functools
import types
import typing as t

from collections import abc as c

from ansible._internal import _display_utils
from ansible._internal._errors import _attribute_unavailable, _captured, _error_factory, _error_utils
from ansible._internal._templating import _engine
from ansible._internal._templating._chain_templar import ChainTemplar
from ansible._internal._worker import _inventory_rpc
from ansible._internal._datatag import _tags
from ansible.errors import AnsibleError, AnsibleTemplateError
from ansible.module_utils._internal._ambient_context import AmbientContextBase
from ansible.module_utils._internal import _dataclass_validation, _event_utils, _messages, _traceback
from ansible.module_utils.datatag import native_type_name, deprecator_from_collection_name, deprecate_value
from ansible.parsing import vault as _vault
from ansible.template import trust_as_template
from ansible.utils.display import Display
from ansible.utils import vars as _vars

if t.TYPE_CHECKING:
    from ansible.playbook.task import Task
    from ansible.inventory.host import Host
    from ansible.executor.task_executor import TaskExecutor

display = Display()

PRESERVE = frozenset(
    {
        '_ansible_no_log',
        'attempts',
        'changed',
        'deprecations',
        'exception',
        'retries',
        'warnings',
    }
)

POLYMORPHIC_RESULT_EXPRESSION = trust_as_template("_task.polymorphic_result")


class NoRecordedResultError(AnsibleError):
    """Error raised when requesting recorded results prior to any being recorded."""

    _default_message = "No task result has been recorded."


class NotALoopError(AnsibleError):
    """Error raised when attempting to access loop state when not in a loop."""

    _default_message = "The current task is not a loop."


@dataclasses.dataclass(kw_only=True, slots=True)
class CurrentTask:
    """
    Instances of this type are exposed to Jinja during register projections and playbook conditional expression eval.

    CAUTION: The shape and behavior of this type is effectively public API.
    """

    def _lazy_transform(self, value: t.Any) -> t.Any:
        """Lazily transform the given value."""
        return TaskContext.current().task_templar.template(value, lazy_options=_engine.LazyOptions.SKIP_TEMPLATES_AND_ACCESS)

    @property
    def result(self) -> c.Mapping[str, object]:
        task_ctx = TaskContext.current()

        try:
            return self._lazy_transform(task_ctx.latest_result)
        except NoRecordedResultError as ex:
            # RPFIX-5: UX: this causes multiple warnings/errors that need to be investigated, some likely from late addition of pre-task `when` projections
            # [WARNING]: An error occurred in a register expression: The _task.result property is unavailable: No task result has been recorded.
            # [WARNING]: An error occurred in a register expression: Error rendering expression: Type 'WarningSummary' is unsupported for variable storage.
            #     - shell: echo {{ item }}
            #       loop: [1,2,3]
            #       failed_when: blar.rc != 0
            #       register:
            #         blar: _task.result
            #
            #     - debug:
            #         var: blar
            # the template engine will turn this into an UndefinedMarker
            raise _attribute_unavailable.AttributeUnavailableError("The _task.result property is unavailable.") from ex

    @property
    def loop_result(self) -> c.Mapping[str, object]:
        task_ctx = TaskContext.current()

        try:
            # in-flight projections bypass the skipped conversions by setting finalize to False
            # RPFIX-5: IMPL: may need to fix Jinja's getattr/item AttributeError handling- if any code invoked beneath one of those raises AttributeError, Jinja
            #  blindly creates a new UndefinedMarker with "CurrentTask has no attr loop_result" and loses the original AttributeError detail. Repro this by
            #  tacking any bogus attr off the end of as_result_dict() here.
            return self._lazy_transform(task_ctx.build_loop_result(preview=True).as_result_dict())
        except NotALoopError as ex:
            raise _attribute_unavailable.AttributeUnavailableError("The _task.loop_result property is unavailable.") from ex

    @property
    def polymorphic_result(self) -> c.Mapping[str, object]:
        task_ctx = TaskContext.current()

        if task_ctx.has_loop_exited:
            return self._lazy_transform(task_ctx.build_loop_result().as_result_dict())

        try:
            return self._lazy_transform(task_ctx.latest_result)
        except NoRecordedResultError as ex:
            raise _attribute_unavailable.AttributeUnavailableError("The _task.polymorphic_result property is unavailable.") from ex


@dataclasses.dataclass(kw_only=True, slots=True)
class PendingChanges:
    """Changes which will be applied when the action completes, including on failure."""

    register_host_variables: dict[VariableLayer, dict[str, object]] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass(kw_only=True)
class TaskContext(AmbientContextBase):
    """Ambient context that wraps task execution on workers. It provides access to the currently executing task."""

    @classmethod
    def create(cls, task: Task, task_vars: dict[str, t.Any], host_name: str) -> t.Self:
        task_vars.update(_task=CurrentTask())

        return cls(
            _task=task,
            _base_task_vars=task_vars,
            _active_task_vars=task_vars,  # starts out as a reference, but becomes a copy of _base_task_vars when starting a loop item
            _host_name=host_name,
        )

    def get_register_projections(self) -> dict[str, _engine.TemplateExpressionWrapper] | None:
        if not self._task.register:
            return None

        return {var_name: _engine.TemplateExpressionWrapper(expression=expression) for var_name, expression in self._task.register.items()}

    @functools.cached_property
    def inventory_rpc_client(self) -> _inventory_rpc.InventoryRPC:
        return _inventory_rpc.InventoryRPC.get_client()

    @property
    def host_name(self) -> str:
        return self._host_name

    @property
    def task(self) -> Task:
        return self._task

    @property
    def task_vars(self) -> dict[str, t.Any]:
        return self._active_task_vars

    @property
    def is_loop(self) -> bool:
        return self._loop_items is not None

    @property
    def has_loop_exited(self) -> bool:
        return self._has_loop_exited

    @property
    def current_item(self) -> object:
        return self._item

    @property
    def current_item_label(self) -> object:
        return self.task.loop_control.label or self.task_templar.variable_name_as_template(self._loop_var)

    _task: Task
    _base_task_vars: dict[str, t.Any]
    _active_task_vars: dict[str, t.Any]
    _host_name: str
    _registered_vars_enabled = False
    _raw_loop_results: list[UnifiedTaskResult] = dataclasses.field(default_factory=list)
    _loop_items: list[object] | None = None
    _has_loop_exited: bool = False
    _loop_var: str | None = None
    _item: object | None = None
    _item_index: int | None = None
    _index_var: str | None = None
    _loop_extended: dict[str, object] | None = None
    _templar: _engine.TemplateEngine | None = None
    _break_when_triggered: bool = False
    _inventory_rpc_client: _inventory_rpc.InventoryRPC | None = None

    pending_changes: PendingChanges | None = None
    """Pending changes which will be applied only if the current task succeeds."""

    @contextlib.contextmanager
    def loop_item_context(self, te: TaskExecutor) -> t.Generator[None]:
        original_task = self.task
        original_play_context = te._play_context

        loop_item_task = self.task.copy(exclude_parent=True, exclude_tasks=True)
        loop_item_task._parent = self.task._parent
        loop_item_play_context = te._play_context.copy()

        self._task = loop_item_task
        te._play_context = loop_item_play_context

        try:
            yield
        finally:
            self._task = original_task
            te._play_context = original_play_context

    def start_loop(self) -> t.Generator[tuple[int, object]]:
        self._loop_var = loop_var = t.cast(str, self.task.loop_control.loop_var)
        self._index_var = index_var = t.cast(str, self.task.loop_control.index_var)

        extended = t.cast(bool, self.task.loop_control.extended)
        extended_allitems = t.cast(bool, self.task.loop_control.extended_allitems)

        items = self._loop_items
        items_len = len(items)

        for item_index, item in enumerate(self._loop_items):
            self._active_task_vars = self._base_task_vars.copy()  # isolate changes to task vars between loop items
            self._templar = None  # we're changing the values used to calculate the templar, null it out so the next requester re-creates it

            self._item = item
            self._item_index = item_index

            loop_vars: dict[str, object] = dict(
                ansible_loop_var=loop_var,
            )

            loop_vars[loop_var] = item

            if index_var:
                loop_vars['ansible_index_var'] = index_var
                loop_vars[index_var] = item_index

            if extended:
                ansible_loop: dict[str, object] = {
                    'index': item_index + 1,
                    'index0': item_index,
                    'first': item_index == 0,
                    'last': item_index + 1 == items_len,
                    'length': items_len,
                    'revindex': items_len - item_index,
                    'revindex0': items_len - item_index - 1,
                }

                if extended_allitems:
                    ansible_loop['allitems'] = items

                try:
                    ansible_loop['nextitem'] = items[item_index + 1]
                except IndexError:
                    pass

                if item_index - 1 >= 0:
                    ansible_loop['previtem'] = items[item_index - 1]

                self._loop_extended = loop_vars['ansible_loop'] = ansible_loop

            self.task_vars.update(loop_vars)

            yield item_index, item

            if self._break_when_triggered:
                break

        self._has_loop_exited = True

    @property
    def task_templar(self) -> _engine.TemplateEngine:
        if not self._templar:
            self._templar = _engine.TemplateEngine(loader=self.task._loader, variables=self.task_vars)

        return self._templar

    @property
    def latest_result(self) -> c.MutableMapping[str, object]:
        if not self._raw_loop_results:
            raise NoRecordedResultError()

        return self._raw_loop_results[-1].as_result_dict()

    def _record_result(self, utr: UnifiedTaskResult) -> None:
        # CAUTION: This method can be called *before* start_loop when validation errors occur.
        #          That results in various instance attributes not being set.
        #          It also means that `self.task.loop_control` may have invalid values.

        self._enable_registered_vars()

        if isinstance(self.task.ignore_errors, bool):
            # HACK: avoid setting to True due to template failures -- this should go away once field attribute templating is fixed
            utr.ignore_errors = self.task.ignore_errors

        if isinstance(self.task.ignore_unreachable, bool):
            # HACK: avoid setting to True due to template failures -- this should go away once field attribute templating is fixed
            utr.ignore_unreachable = self.task.ignore_unreachable

        # _item_index will be None if start_loop was not called due to a field attribute error
        if not TaskContext.current().is_loop or self._item_index is None:
            self._raw_loop_results.clear()
            self._raw_loop_results.append(utr)

            return

        if TaskContext.current().has_loop_exited:
            return

        # now update the result with the item info, and append the result
        # to the list of results
        utr.loop_item = self._item
        utr.loop_var = self._loop_var
        utr.loop_extended = self._loop_extended

        if self._index_var:
            utr.loop_index = self._item_index
            utr.loop_index_var = self._index_var

        item_index = self._item_index
        result_count = len(self._raw_loop_results)

        if item_index == result_count:
            self._raw_loop_results.append(utr)  # add new result
        elif item_index + 1 == result_count:
            self._raw_loop_results[item_index] = utr  # replace existing result
        else:
            raise RuntimeError(f'Item index {item_index} does not match {result_count}.')

        self._populate_item_label(utr)

    def _populate_item_label(self, utr: UnifiedTaskResult) -> None:
        # gets templated here unlike rest of loop_control fields, depends on loop_var above
        item_label: object = ...

        if item_label_template := self.task.loop_control.label:
            try:
                item_label = self.task_templar.template(item_label_template)
            except AnsibleTemplateError as ex:
                display.error_as_warning('Failed to template loop_control label.', ex, obj=item_label_template)

        if item_label is ...:
            if self._loop_var is None:
                # _loop_var will be None if start_loop was not called due to a field attribute error (even if the error is for another field)
                item_label = ''
            else:
                item_label = self.task_templar.resolve_variable_expression(self._loop_var)

        utr.loop_item_label = item_label

    def build_loop_result(self, preview: bool = False) -> UnifiedTaskResult:
        if not self.is_loop:
            raise NotALoopError()

        if not preview and (self._item_index is None or len(self._raw_loop_results) not in (self._item_index, self._item_index + 1)):
            # RPFIX-9: FUTURE: can we ditch preview while retaining this safety check?
            # Loop results can be queried before or after loop results are recorded, so we need to accept a range of results.
            raise RuntimeError(f"Mismatch between item index {self._item_index} and loop result count {len(self._raw_loop_results)}.")

        # create the overall result item
        utr = UnifiedTaskResult.from_action_result_dict()
        utr.loop_results = self._raw_loop_results

        # RPFIX-5: IMPL: all the fields set in this loop could be converted to properties
        for item in self._raw_loop_results:
            if item.no_log:
                utr.no_log = True  # ensure no_log processing recognizes at least one item needs to be censored

            utr._extend_warnings(item.warnings)
            utr._extend_deprecations(item.deprecations)

        if all(item.skipped for item in self._raw_loop_results):
            utr.set_skipped()
            utr.msg = 'All items skipped'
        elif utr.failed:
            utr.msg = 'One or more items failed'
        else:
            utr.msg = 'All items completed'

        return utr

    def update_task_vars(self, variables: dict[str, t.Any]) -> None:
        """Update task variables for both the active task and, in the case of loops, subsequent tasks."""
        for key, value in variables.items():
            self._base_task_vars[key] = self._active_task_vars[key] = value

    def _enable_registered_vars(self) -> None:
        """
        Inject registered variable expressions into task vars.
        This needs to be done one time after the action handler runs, or if a task results in no action running.
        Unlike other modifications to task vars, these changes will be persisted between loop items.
        """
        if self._registered_vars_enabled:
            return

        if register_projections := self.get_register_projections():
            self.update_task_vars(register_projections)

        self._registered_vars_enabled = True


TaskArgsFinalizerCallback = t.Callable[[str, t.Any, _engine.TemplateEngine, t.Any], t.Any]
"""Type alias for the shape of the `ActionBase.finalize_task_arg` method."""


class TaskArgsChainTemplar(ChainTemplar):
    """
    A ChainTemplar that carries a user-provided context object, optionally provided by `ActionBase.get_finalize_task_args_context`.
    TaskArgsFinalizer provides the context to each `ActionBase.finalize_task_arg` call to allow for more complex/stateful customization.
    """

    def __init__(self, *sources: c.Mapping, templar: _engine.TemplateEngine, callback: TaskArgsFinalizerCallback, context: t.Any) -> None:
        super().__init__(*sources, templar=templar)

        self.callback = callback
        self.context = context

    def template(self, key: t.Any, value: t.Any) -> t.Any:
        return self.callback(key, value, self.templar, self.context)


class TaskArgsFinalizer:
    """Invoked during task args finalization; allows actions to override default arg processing (e.g., templating)."""

    def __init__(self, *args: c.Mapping[str, t.Any] | str | None, templar: _engine.TemplateEngine) -> None:
        self._args_layers = [arg for arg in args if arg is not None]
        self._templar = templar

    def finalize(self, callback: TaskArgsFinalizerCallback, context: t.Any) -> dict[str, t.Any]:
        from ansible import constants

        resolved_layers: list[c.Mapping[str, t.Any]] = []

        for layer in self._args_layers:
            if isinstance(layer, (str, _vault.EncryptedString)):  # EncryptedString can hide a template
                if constants.config.get_config_value('INJECT_FACTS_AS_VARS'):
                    display.warning(
                        "Using a template for task args is unsafe in some situations "
                        "(see https://docs.ansible.com/ansible/devel/reference_appendices/faq.html#argsplat-unsafe).",
                        obj=layer,
                    )

                resolved_layer = self._templar.resolve_to_container(layer, options=_engine.TemplateOptions(value_for_omit={}))
            else:
                resolved_layer = layer

            if not isinstance(resolved_layer, dict):
                raise AnsibleError(f'Task args must resolve to a {native_type_name(dict)!r} not {native_type_name(resolved_layer)!r}.', obj=layer)

            resolved_layers.append(resolved_layer)

        ct = TaskArgsChainTemplar(*reversed(resolved_layers), templar=self._templar, callback=callback, context=context)

        return ct.as_dict()


class Source(enum.Enum):
    ACTION = enum.auto()
    ANY = enum.auto()


class Destination(enum.Enum):
    CALLBACK = enum.auto()
    NOT_CALLBACK = enum.auto()
    ANY = enum.auto()


@dataclasses.dataclass(kw_only=True, frozen=True, slots=True)
class FieldSettings:
    key: str | None = None
    source: Source | None = None
    destination: Destination | None = None
    conversion_func: t.Callable[[object], object] | None = None


_type = type


@dataclasses.dataclass(kw_only=True, frozen=True, slots=True)
class ResolvedField:
    name: str
    type: type
    optional: bool
    field: dataclasses.Field | None
    metadata: FieldSettings

    @classmethod
    def from_field(cls, type_hints: dict[str, t.Any], metadata: FieldSettings, dc_field: dataclasses.Field) -> t.Self:
        resolved_type = type_hints[dc_field.name]

        if isinstance(resolved_type, types.UnionType):
            args = resolved_type.__args__

            if len(args) == 2 and args[0] is types.NoneType:
                resolved_type = args[1]
                optional = True
            elif len(args) == 2 and args[1] is types.NoneType:
                resolved_type = args[0]
                optional = True
            else:
                raise NotImplementedError(f"Unexpected union type args: {args}")
        else:
            optional = False

        if isinstance(resolved_type, types.GenericAlias):
            resolved_type = t.get_origin(resolved_type)
        elif (orig_bases := types.get_original_bases(resolved_type)) and orig_bases[0] is t.TypedDict:
            resolved_type = dict

        return cls(
            name=dc_field.name,
            type=resolved_type,
            optional=optional,
            field=dc_field,
            metadata=metadata,
        )

    @property
    def result_key(self) -> str:
        return self.metadata.key or self.name


def export_only(
    key: str | None = None,
    *,
    destination: Destination = Destination.ANY,
    conversion_func: t.Callable[[object], object] | None = None,
) -> dict[str, t.Any]:
    return field(
        key,
        source=None,
        destination=destination,
        conversion_func=conversion_func,
    )


def import_export(
    key: str | None = None,
    *,
    source: Source = Source.ANY,
    destination: Destination = Destination.ANY,
    conversion_func: t.Callable[[object], object] | None = None,
) -> dict[str, t.Any]:
    return field(
        key,
        source=source,
        destination=destination,
        conversion_func=conversion_func,
    )


def field(
    key: str | None = None,
    *,
    source: Source | None = None,
    destination: Destination | None = None,
    conversion_func: t.Callable[[object], object] | None = None,
) -> dict[str, t.Any]:
    return dict(
        ansible=FieldSettings(
            key=key,
            source=source,
            destination=destination,
            conversion_func=conversion_func,
        )
    )


_DEFAULT_FIELD_SETTINGS = FieldSettings()


def _convert_no_log(value: object) -> bool:
    if not isinstance(value, bool):
        # RPFIX-5: UX: do we want to keep any kind of custom warning regarding this value always being True when not a bool?
        # display.warning(f'Invalid _ansible_no_log value of type {type(value).__name__!r} in task result, output will be masked.')
        value = True

    return value


@dataclasses.dataclass(kw_only=True, frozen=True, slots=True)
class AddHost:
    host_name: str
    host_vars: dict[str, object] | None = None
    parent_group_names: list[str] | None = None

    def __post_init__(self): ...

    @classmethod
    def from_dict(cls, value: object) -> AddHost | None:
        if isinstance(value, dict):
            if groups := value.pop('groups'):
                value['parent_group_names'] = groups

            try:
                return cls(**value)
            except Exception as ex:
                display.error_as_warning('Ignoring invalid add_host value in task result.', exception=ex)

        return None


_dataclass_validation.inject_post_init_validation(AddHost, allow_subclasses=True)


@dataclasses.dataclass(kw_only=True, frozen=True, slots=True)
class AddGroup:
    group_name: str
    parent_group_names: list[str]

    def __post_init__(self): ...


_dataclass_validation.inject_post_init_validation(AddGroup, allow_subclasses=True)


class StatsDict(t.TypedDict):
    data: dict[str, object]
    per_host: t.NotRequired[object | None]  # really bool, but no validation
    aggregate: t.NotRequired[object | None]  # really bool, but no validation


def _convert_stats(value: t.Any) -> StatsDict | None:
    if not isinstance(value, dict) or not isinstance(value.get('data'), dict):
        display.warning('Ignoring invalid ansible_stats value in task result.')
        value = None

    return value


def _convert_warnings(value: object) -> list[_messages.WarningSummary] | None:
    if not value:
        return None

    if not isinstance(value, list):
        display.warning(f"Task result `warnings` was {native_type_name(value)} instead of {native_type_name(list)} and has been discarded.")
        return None

    warnings: list[_messages.WarningSummary] = []

    for warning in value:
        if not isinstance(warning, _messages.WarningSummary):
            # translate non-WarningMessageDetail messages
            warning = _messages.WarningSummary(
                event=_messages.Event(
                    msg=str(warning),
                ),
            )

        if warning_ctx := _display_utils.DeferredWarningContext.current(optional=True):
            warning_ctx.capture(warning)
        else:
            warnings.append(warning)

    return warnings


def _convert_deprecations(value: object) -> list[_messages.DeprecationSummary] | None:
    if not value:
        return None

    if not isinstance(value, list):
        display.warning(f"Task result `deprecations` was {native_type_name(value)} instead of {native_type_name(list)} and has been discarded.")
        return None

    deprecations: list[_messages.DeprecationSummary] = []

    for deprecation in value:
        if not isinstance(deprecation, _messages.DeprecationSummary):
            # translate non-DeprecationSummary message dicts
            try:
                if (collection_name := deprecation.pop('collection_name', ...)) is not ...:
                    # deprecated: description='enable the deprecation message for collection_name' core_version='2.23'
                    # CAUTION: This deprecation cannot be enabled until the replacement (deprecator) has been documented, and the schema finalized.
                    # self.deprecated('The `collection_name` key in the `deprecations` dictionary is deprecated.', version='2.27')
                    deprecation.update(deprecator=deprecator_from_collection_name(collection_name))

                deprecation = _messages.DeprecationSummary(
                    event=_messages.Event(
                        msg=deprecation.pop('msg'),
                    ),
                    **deprecation,
                )
            except Exception as ex:
                display.error_as_warning("Task result `deprecations` contained an invalid item.", exception=ex)

        if warning_ctx := _display_utils.DeferredWarningContext.current(optional=True):
            warning_ctx.capture(deprecation)
        else:
            deprecations.append(deprecation)

    return deprecations


DROP_AND_WARN = ResolvedField(name='__drop__', type=str, optional=True, field=None, metadata=FieldSettings())


class VariableLayer(enum.IntEnum):
    """
    Variable layer at which variables are registered.

    CAUTION: This enum is exposed as public API in ActionBase.
    """

    # IMPORTANT: The order of these enum values determines the order in which they will be applied to variable manager.
    #            This matters for precedence of layers for which variable manager does not independently track.

    CACHEABLE_FACT = enum.auto()
    INCLUDE_VARS = enum.auto()
    EPHEMERAL_FACT = enum.auto()
    REGISTER_VARS = enum.auto()


@dataclasses.dataclass(kw_only=True, slots=True)
class UnifiedTaskResult:
    _changed: bool | None = dataclasses.field(default=None, metadata=import_export("changed"))

    @property
    def changed(self) -> bool:
        if self._changed is not None or self.loop_results is None:
            return bool(self._changed)

        return any(loop_result.changed for loop_result in self.loop_results)

    @changed.setter
    def changed(self, value: bool) -> None:
        self._changed = value

    _failed: bool | None = dataclasses.field(default=None, metadata=import_export("failed", destination=Destination.NOT_CALLBACK))

    @property
    def failed(self) -> bool:
        if self._failed is None and self.rc is not None and self.rc != 0:
            return True

        if self.failed_when_result is not None:
            return bool(self.failed_when_result)

        if self.loop_results and any(loop_result.failed_when_result is not None for loop_result in self.loop_results):
            return any(loop_result.failed_when_result for loop_result in self.loop_results)

        if self._failed is not None or self.loop_results is None:
            return bool(self._failed)

        return any(loop_result.failed for loop_result in self.loop_results)

    @failed.setter
    def failed(self, value: bool) -> None:
        self._failed = value

    _unreachable: bool | None = dataclasses.field(default=None, metadata=import_export("unreachable"))

    @property
    def unreachable(self) -> bool | None:
        if self._unreachable is not None or self.loop_results is None:
            result = bool(self._unreachable)
        else:
            result = any(loop_result.unreachable for loop_result in self.loop_results)

        return result or None

    @unreachable.setter
    def unreachable(self, value: bool) -> None:
        self._unreachable = value

    _skipped: bool | None = dataclasses.field(default=None, metadata=import_export("skipped", destination=Destination.NOT_CALLBACK))

    @property
    def skipped(self) -> bool | None:
        if self._skipped is not None or self.loop_results is None:
            result = bool(self._skipped)
        else:
            # Loop tasks are only considered skipped if all items were skipped.
            result = all(loop_result.skipped for loop_result in self.loop_results)

        return result or None

    @skipped.setter
    def skipped(self, value: bool) -> None:
        self._skipped = value

    _ignore_errors: bool = False  # formerly _ansible_ignore_errors

    @property
    def ignore_errors(self) -> bool:
        return self._ignore_errors if self.loop_results is None else any(loop_result._ignore_errors for loop_result in self.loop_results)

    @ignore_errors.setter
    def ignore_errors(self, value: bool) -> None:
        self._ignore_errors = value

    _ignore_unreachable: bool = False  # formerly _ansible_ignore_unreachable

    @property
    def ignore_unreachable(self) -> bool:
        return self._ignore_unreachable if self.loop_results is None else any(loop_result._ignore_unreachable for loop_result in self.loop_results)

    @ignore_unreachable.setter
    def ignore_unreachable(self, value: bool) -> None:
        self._ignore_unreachable = value

    _warnings: list[_messages.WarningSummary] | None = dataclasses.field(
        default=None,
        metadata=import_export("warnings", conversion_func=_convert_warnings),
    )

    @property
    def warnings(self) -> list[_messages.WarningSummary] | None:
        if (current_warnings := self._warnings) is None and (warning_ctx := _display_utils.DeferredWarningContext.current(optional=True)):
            current_warnings = warning_ctx.get_warnings()

        return current_warnings

    _deprecations: list[_messages.DeprecationSummary] | None = dataclasses.field(
        default=None,
        metadata=import_export("deprecations", conversion_func=_convert_deprecations),
    )

    @property
    def deprecations(self) -> list[_messages.DeprecationSummary] | None:
        if (current_deprecations := self._deprecations) is None and (warning_ctx := _display_utils.DeferredWarningContext.current(optional=True)):
            current_deprecations = warning_ctx.get_deprecation_warnings()

        return current_deprecations

    ansible_facts: dict[str, t.Any] | None = dataclasses.field(default=None, metadata=import_export())
    async_result: dict[str, t.Any] | None = dataclasses.field(default=None, metadata=export_only())
    ansible_parsed: bool | None = None  # formerly _ansible_parsed
    exception: _messages.ErrorSummary | None = dataclasses.field(default=None, metadata=import_export())
    finished: bool | None = dataclasses.field(default=None, metadata=import_export())
    msg: object | None = dataclasses.field(default=None, metadata=import_export())
    no_log: bool | None = dataclasses.field(
        default=None, metadata=import_export("_ansible_no_log", conversion_func=_convert_no_log, source=Source.ACTION, destination=Destination.CALLBACK)
    )
    verbose_always: bool | None = dataclasses.field(
        default=None, metadata=import_export("_ansible_verbose_always", source=Source.ACTION, destination=Destination.CALLBACK)
    )
    verbose_override: bool | None = dataclasses.field(
        default=None, metadata=import_export("_ansible_verbose_override", source=Source.ACTION, destination=Destination.CALLBACK)
    )
    rc: int | None = dataclasses.field(default=None, metadata=import_export())
    suppress_tmpdir_delete: bool | None = dataclasses.field(default=None, metadata=field("_ansible_suppress_tmpdir_delete", source=Source.ANY))
    module_stderr: str | None = dataclasses.field(default=None, metadata=export_only())
    module_stdout: str | None = dataclasses.field(default=None, metadata=export_only())
    result_data: dict[str, object] = dataclasses.field(default_factory=dict)
    is_module: bool
    attempts: int | None = dataclasses.field(default=None, metadata=export_only())
    retries: int | None = dataclasses.field(default=None, metadata=export_only())
    notify: list[str] = dataclasses.field(default_factory=list)
    delegated_host: str | None = None
    callback_delegated_vars_subset: dict[str, object] | None = dataclasses.field(
        default=None, metadata=export_only("_ansible_delegated_vars", destination=Destination.CALLBACK)
    )
    diff: object | None = dataclasses.field(default=None, metadata=import_export())  # RPFIX-9: FUTURE: validation with custom conversion func
    pending_changes: PendingChanges = dataclasses.field(default_factory=PendingChanges)
    """Changes which will be applied when the action completes, including on failure."""
    stats: StatsDict | None = dataclasses.field(default=None, metadata=import_export('ansible_stats', conversion_func=_convert_stats))
    async_job_id: str | None = dataclasses.field(default=None, metadata=import_export('ansible_job_id'))
    include_file: str | None = None
    include_args: dict[str, object] | None = None  # RPFIX-9: FUTURE: make this a dataclass
    registered_values: c.Mapping[str, object] | None = None
    """Values to register unconditionally, including on failure or when skipped."""

    skip_reason: str | None = dataclasses.field(default=None, metadata=export_only())
    skipped_reason: str | None = dataclasses.field(default=None, metadata=export_only())
    """The `skipped_reason` field is deprecated. Use `skip_reason` instead."""

    false_condition: object | None = dataclasses.field(default=None, metadata=export_only())

    changed_when_result: object | None = dataclasses.field(default=None, metadata=export_only())  # RPFIX-9: FUTURE: `bool | str` once validator supports that
    failed_when_result: object | None = dataclasses.field(default=None, metadata=export_only())  # RPFIX-9: FUTURE: `bool | str` once validator supports that
    break_when_result: object | None = dataclasses.field(default=None, metadata=export_only())  # RPFIX-9: FUTURE: `bool | str` once validator supports that

    changed_when_suppressed_exception: _messages.ErrorSummary | None = dataclasses.field(default=None, metadata=export_only())
    failed_when_suppressed_exception: _messages.ErrorSummary | None = dataclasses.field(default=None, metadata=export_only())
    break_when_suppressed_exception: _messages.ErrorSummary | None = dataclasses.field(default=None, metadata=export_only())

    loop_var: str | None = dataclasses.field(default=None, metadata=export_only('ansible_loop_var'))
    loop_index_var: str | None = dataclasses.field(default=None, metadata=export_only('ansible_index_var'))
    loop_item_label: object | None = dataclasses.field(default=None, metadata=export_only('_ansible_item_label', destination=Destination.CALLBACK))
    loop_item: object | None = dataclasses.field(default=None, metadata=export_only())  # exported as self.loop_var value
    loop_index: object | None = dataclasses.field(default=None, metadata=export_only())  # exported as self.loop_index_var value
    loop_extended: dict[str, object] | None = dataclasses.field(default=None, metadata=export_only('ansible_loop'))
    loop_results: list[t.Self] | None = None

    @classmethod
    @functools.cache
    def get_result_key_to_resolved_field_mapping(cls, source_is_module: bool) -> dict[str, ResolvedField]:
        mapping = {}
        type_hints = t.get_type_hints(cls)

        for dc_field in dataclasses.fields(cls):
            metadata = t.cast(FieldSettings, dc_field.metadata.get('ansible', _DEFAULT_FIELD_SETTINGS))
            resolved_field = ResolvedField.from_field(type_hints, metadata, dc_field)

            if metadata.source is Source.ACTION and source_is_module:
                mapping[resolved_field.result_key] = DROP_AND_WARN
            elif metadata.source is Source.ACTION or metadata.source is Source.ANY:
                mapping[resolved_field.result_key] = resolved_field

        return mapping

    @classmethod
    @functools.cache
    def _get_field_name_to_result_key_mapping(cls, *, for_callback: bool, for_round_trip: bool) -> dict[str, str]:
        mapping = {}

        for dc_field in dataclasses.fields(cls):
            metadata = t.cast(FieldSettings, dc_field.metadata.get('ansible', _DEFAULT_FIELD_SETTINGS))

            if (
                metadata.destination is Destination.ANY
                or ((for_round_trip or for_callback) and metadata.destination is Destination.CALLBACK)
                or ((for_round_trip or not for_callback) and metadata.destination is Destination.NOT_CALLBACK)
            ):
                key = metadata.key or dc_field.name
                field_name = key if hasattr(cls, key) else dc_field.name  # properties matching the exported key take precedence over the field during export

                mapping[field_name] = key

        return mapping

    @classmethod
    def from_action_result_dict(cls, result: dict[str, object] | None = None) -> t.Self:
        return cls._from_result_dict(result or {}, source_is_module=False)

    @classmethod
    def _from_result_dict(cls, result: dict[str, object], source_is_module: bool) -> t.Self:
        if not isinstance(result, dict):
            raise TypeError(f'Malformed result. Received {type(result)} instead of {dict}.')

        fields = cls.get_result_key_to_resolved_field_mapping(source_is_module=source_is_module)
        result_data: dict[str, object] = {}
        kwargs: dict[str, t.Any] = dict(
            result_data=result_data,
        )

        for key, value in result.items():
            if resolved_field := fields.get(key):
                if resolved_field is DROP_AND_WARN:
                    display.warning(f"Removed reserved key {key!r} from module result.", obj=value)
                else:
                    kwargs[resolved_field.name] = cls.convert_field(resolved_field, value)
            elif source_is_module and key.startswith('_ansible_'):
                display.warning(f"Removed reserved key {key!r} from module result.", obj=value)
            else:
                result_data[key] = value

        return cls(is_module=source_is_module, **kwargs)

    def _result_key_magic(self, key: str) -> str:
        match key:
            case "loop_item":
                return self.loop_var or 'item'  # RPFIX-9: FUTURE: centralize the default logic

            case "loop_index":
                return self.loop_index_var

            case _:
                return key

    def as_result_dict(self, *, for_callback: bool = False, for_round_trip: bool = False, censor_callback_result: bool = False) -> dict[str, object]:
        result: dict[str, t.Any] = {
            self._result_key_magic(result_key): value
            for field_name, result_key in self._get_field_name_to_result_key_mapping(for_callback=for_callback, for_round_trip=for_round_trip).items()
            if (value := getattr(self, field_name)) is not None
        }

        result.update(self.result_data)

        # RPFIX-5: IMPL: is this where we want stdout/stderr handling?
        # pre-split stdout/stderr into lines if needed

        if 'stdout' in result and 'stdout_lines' not in result:
            # if the value is 'False', a default won't catch it.
            txt = result.get('stdout', None) or ''
            result.update(stdout_lines=txt.splitlines())

        if 'stderr' in result and 'stderr_lines' not in result:
            # if the value is 'False', a default won't catch it.
            txt = result.get('stderr', None) or ''
            result.update(stderr_lines=txt.splitlines())

        if for_callback:
            if censor_callback_result:
                result = {key: value for key in PRESERVE if (value := result.get(key, ...)) is not ...}
                result.update(censored="the output has been hidden due to the fact that 'no_log: true' was specified for this result")

        if self.loop_results is not None:
            # loop results need to be added after censor_result on the outer result since it's currently naive about whether it's looking at a loop item or not
            result.update(
                results=[
                    loop_result.as_result_dict(
                        for_callback=for_callback,
                        for_round_trip=for_round_trip,
                        censor_callback_result=censor_callback_result or loop_result.no_log,
                    )
                    for loop_result in self.loop_results
                ]
            )

        if for_callback:
            result = _vars.transform_to_native_types(result)

        return result

    def finalize_registered_values(self) -> None:
        task_ctx = TaskContext.current()

        if not (register_projections := task_ctx.get_register_projections()):
            return

        registered_values = {}
        registered_errors = []

        for var_name, expression in register_projections.items():
            try:
                registered_values[var_name] = task_ctx.task_templar.template(expression)
            except Exception as ex:
                event = _error_factory.ControllerEventFactory.from_exception(ex, False)
                event_message = _event_utils.format_event_brief_message(event)
                undef_message = f"The variable {var_name!r} is undefined because its register expression failed: {event_message}"
                undef_value = trust_as_template(f"{{{{ undef({undef_message!r}) }}}}")

                if expression_origin := _tags.Origin.get_tag(expression.expression):
                    undef_value = expression_origin.tag(undef_value)

                registered_values[var_name] = undef_value

                try:
                    raise Exception(f"Register projection {var_name!r} failed.") from ex
                except Exception as ex:
                    registered_errors.append(ex)

        # RPFIX-9: FUTURE: merge registered_values into the pending VariableLayer.REGISTER_VARS layer instead of having a separate field
        self.registered_values = registered_values

        if not registered_errors:
            return

        chain = (
            _messages.EventChain(
                msg_reason="The original task error before register failed was:",
                traceback_reason="The above exception occurred before the following exception:",
                event=self.exception.event,
            )
            if self.exception
            else None
        )

        try:
            raise ExceptionGroup(
                f"Task failed due to errors in {len(registered_errors)} out of {len(register_projections)} register projections.",
                registered_errors,
            )
        except ExceptionGroup as ex:
            event = _error_factory.ControllerEventFactory.from_exception(ex, _traceback.is_traceback_enabled(_traceback.TracebackEvent.ERROR))
            event = dataclasses.replace(event, chain=chain)

            self.failed = True
            self.exception = _messages.ErrorSummary(
                event=event,
            )

    def _extend_warnings(self, warnings: c.Iterable[_messages.WarningSummary] | None) -> None:
        if not warnings:
            return

        if self._warnings is None:
            self._warnings = []

        self._warnings.extend(warnings)

    def _extend_deprecations(self, deprecations: c.Iterable[_messages.DeprecationSummary]) -> None:
        if not deprecations:
            return

        if self._deprecations is None:
            self._deprecations = []

        self._deprecations.extend(deprecations)

    @classmethod
    def convert_field(cls, resolved_field: ResolvedField, value: object) -> object:
        if value is None:
            return None if resolved_field.optional else resolved_field.field.default

        if resolved_field.type is _messages.ErrorSummary:
            return value  # RPFIX-5: VALIDATION: defer conversion to __post_init__

        # RPFIX-5: VALIDATION: this doesn't handle special types like exception which require extra conversion

        if not resolved_field.metadata.conversion_func and isinstance(value, resolved_field.type):
            return value

        # RPFIX-5: VALIDATION: this type checking doesn't validate types within containers (mapping, dict, etc.)

        help_text = f"Values for result key {resolved_field.result_key!r} must be of type {native_type_name(resolved_field.type)}."

        conversion_func = resolved_field.metadata.conversion_func or resolved_field.type

        try:
            result = conversion_func(value)
        except Exception:
            result = resolved_field.field.default
            display.warning(
                f'Value for result key {resolved_field.result_key!r} of type {native_type_name(value)} was replaced with {result}.',
                obj=value,
                help_text=help_text,
            )
        else:
            if not isinstance(value, resolved_field.type):
                # RPFIX-5: UX: this will still give duplicate warnings -- probably just need to be explicit and have conversion funcs do warnings
                display.warning(
                    msg=f'Value for result key {resolved_field.result_key!r} of type {native_type_name(value)} '
                    f'was converted to {native_type_name(resolved_field.type)}.',
                    obj=value,
                    help_text=help_text,
                )

        return result

    def __post_init__(self) -> None:
        # Normalize the result `exception`, if any, to be a `CapturedErrorSummary` instance.
        # If a new `CapturedErrorSummary` was created, the `error_type` will be `cls`.
        # The `exception` key will be removed if falsey.
        # A `CapturedErrorSummary` instance will be returned if `failed` is truthy.
        if isinstance(self.exception, _captured.CapturedErrorSummary):
            self.is_module = self.exception.is_module
        elif isinstance(self.exception, _messages.ErrorSummary):
            self.exception = _captured.CapturedErrorSummary(
                event=self.exception.event,
                error_context=self._captured_error_context,
                error_message=self._captured_error_message,
                is_module=self.is_module,
            )
        elif self.failed or self.exception:
            # RPFIX-5: UX: warn/deprecate if exception is present without failed being True (and do what with failed?)

            # translate non-ErrorSummary errors
            self.exception = _captured.CapturedErrorSummary(
                event=_messages.Event(
                    msg=str(self.msg or 'Unknown error.'),
                    formatted_traceback=_normalize_traceback(self.exception),
                ),
                error_context=self._captured_error_context,
                error_message=self._captured_error_message,
                is_module=self.is_module,
            )
        else:
            self.exception = None

    @classmethod
    def from_module_result_dict(cls, result: dict[str, object]) -> t.Self:
        if (results := result.get('results', ...)) is not ... and (not isinstance(results, c.Sequence) or isinstance(results, str)):
            # deprecated: description='deprecate the value of ansible_module_results' core_version='2.25'
            # results = deprecate_value(
            #     value=results,
            #     msg="The 'ansible_module_results' result key is deprecated.",
            #     help_text="Use the 'results' result key instead.",
            #     version="2.29",
            # )

            result['ansible_module_results'] = results

        return cls._from_result_dict(result, source_is_module=True)

    @staticmethod
    def create_from_module_exception(exception: BaseException, accept_result_contribution: bool = False) -> UnifiedTaskResult:
        """Return a failed task result dict from the given exception."""
        return UnifiedTaskResult._create_from_exception(exception, accept_result_contribution=accept_result_contribution, source_is_module=True)

    @staticmethod
    def _create_from_exception(exception: BaseException, *, accept_result_contribution: bool = False, source_is_module: bool) -> UnifiedTaskResult:
        """Return a failed task result dict from the given exception."""
        event = _error_factory.ControllerEventFactory.from_exception(exception, _traceback.is_traceback_enabled(_traceback.TracebackEvent.ERROR))

        omit_failed_key = False
        omit_exception_key = False
        utr = UnifiedTaskResult(is_module=source_is_module)

        if accept_result_contribution:
            while exception:
                if isinstance(exception, _error_utils.ContributesToTaskResult):
                    utr = exception.as_task_result(utr)
                    omit_failed_key = exception.omit_failed_key
                    omit_exception_key = exception.omit_exception_key
                    break

                exception = exception.__cause__

        if omit_failed_key:
            utr.failed = None
        else:
            utr.failed = True

        if omit_exception_key:
            utr.exception = None
        else:
            utr.exception = _messages.ErrorSummary(event=event)

        if not utr.msg:
            # if nothing contributed `msg`, generate one from the exception messages
            utr.msg = _event_utils.format_event_brief_message(event)

        return utr

    def finalize_warnings(self, warning_ctx: _display_utils.DeferredWarningContext) -> None:
        if warnings := warning_ctx.get_warnings():
            self._warnings = warnings

        if deprecation_warnings := warning_ctx.get_deprecation_warnings():
            self._deprecations = deprecation_warnings

    @property
    def _captured_error_message(self) -> str:
        return 'Module failed.' if self.is_module else 'Action failed.'

    @property
    def _captured_error_context(self) -> str:
        return 'target' if self.is_module else 'action'

    def set_fact(self, key: str, value: object) -> None:
        if self.ansible_facts is None:
            self.ansible_facts = {}

        self.ansible_facts[key] = value

    def remove_internal_keys(self):
        # cleanse fact values that are allowed from actions but not modules
        if self.ansible_facts:
            for key in list(self.ansible_facts):
                # RPFIX-5: IMPL: move this into the control plane and inject into the results as needed
                if key.startswith('discovered_interpreter_') or key.startswith('ansible_discovered_interpreter_'):
                    self.ansible_facts.pop(key)

    @staticmethod
    def create_from_action_exception(exception: BaseException, accept_result_contribution: bool = False) -> UnifiedTaskResult:
        """Return a failed task result dict from the given exception."""
        utr = UnifiedTaskResult._create_from_exception(exception, accept_result_contribution=accept_result_contribution, source_is_module=False)

        TaskContext.current()._record_result(utr)

        return utr

    @classmethod
    @contextlib.contextmanager
    def create_and_record(cls, result: dict[str, t.Any] | None = None) -> t.Generator[t.Self]:
        utr = cls.from_action_result_dict(result or {})
        yield utr
        TaskContext.current()._record_result(utr)

    def maybe_raise_on_result(self) -> None:
        """Raise an exception if the result indicated failure."""
        if self.exception and self.failed:  # even though error detail was normalized, only raise it if the result indicated failure
            raise _captured.AnsibleResultCapturedError(self.exception.event, self)

    def set_skipped(self, reason: str | None = None, include_skipped_reason: bool = False) -> None:
        self.changed = False
        self.skipped = True

        if reason:
            self.skip_reason = reason

        if include_skipped_reason:
            self.skipped_reason = deprecate_value(
                value=reason,
                msg="The 'skipped_reason' value is deprecated.",
                help_text="Use 'skip_reason' instead.",
                version="2.24",
            )

    @classmethod
    def record_conditional_false(cls, conditional_expression: object) -> t.Self:
        with cls.create_and_record() as utr:
            utr.set_skipped('Conditional result was False')
            utr.false_condition = conditional_expression

        return utr

    def set_break_when_result(self, value: bool | AnsibleError) -> None:
        task_ctx = TaskContext.current()

        result: bool | str

        if isinstance(value, bool):
            result = value
            break_when = value
        else:
            result = str(value)

            try:
                # The original error's obj should contain the failed expression, which should be origin tagged.
                # This error is intended to provide additional context (the message) for the original error.
                # By repeating the original error's obj here, the error display will group the two errors together.
                raise AnsibleError("A 'break_when' expression failed.", obj=value.obj) from value
            except Exception as ex:
                wrapped_ex = ex

            break_when = True

            if self.exception:
                self.break_when_suppressed_exception = self.exception

            self.failed = True
            self.exception = _messages.ErrorSummary(
                # pylint: disable=used-before-assignment
                event=_error_factory.ControllerEventFactory.from_exception(wrapped_ex, _traceback.is_traceback_enabled(_traceback.TracebackEvent.ERROR)),
            )

        if break_when:
            task_ctx._break_when_triggered = True
            task_ctx._has_loop_exited = True

        self.break_when_result = result

    def set_changed_when_result(self, value: bool | AnsibleError) -> None:
        result: bool | str

        if isinstance(value, bool):
            result = value

            self.changed = value
        else:
            if self.exception:
                self.changed_when_suppressed_exception = self.exception

            result = str(value)

            try:
                # The original error's obj should contain the failed expression, which should be origin tagged.
                # This error is intended to provide additional context (the message) for the original error.
                # By repeating the original error's obj here, the error display will group the two errors together.
                raise AnsibleError("A 'changed_when' expression failed.", obj=value.obj) from value
            except Exception as ex:
                wrapped_ex = ex

            self.failed = True
            self.exception = _messages.ErrorSummary(
                # pylint: disable=used-before-assignment
                event=_error_factory.ControllerEventFactory.from_exception(wrapped_ex, _traceback.is_traceback_enabled(_traceback.TracebackEvent.ERROR)),
            )

        self.changed_when_result = result

    def set_failed_when_result(self, value: bool | AnsibleError) -> None:
        # RPFIX-5: DOC: document exception suppression logic in changelog et al
        #  The use of any failed_when expression will suppress the existing value of `exception` and store it in `failed_when_suppressed_exception`,
        #  regardless if the expression result is True, False, or an exception. For True and exception, the `exception` value in the task result becomes
        #  one generated (or caused) by failed_when.

        if self.exception:
            self.failed_when_suppressed_exception = self.exception
            self.exception = None

        result: bool | str

        if isinstance(value, bool):
            result = value

            self.failed = value

            if value:
                event = _messages.Event(
                    msg="A 'failed_when' expression evaluated to 'True'.",
                    # RPFIX-5: UX: We should be able to show more useful information here than just a static message,
                    #  including any shadowed msg we may have had and the (origin-tagged) failed_when expression itself.
                    # chain=None if self.msg is None else _messages.EventChain(
                    #     msg_reason='DO THINGS HERE',
                    #     traceback_reason='DO THINGS HERE',
                    #     event=_messages.Event(
                    #         msg=str(self.msg),
                    #         help_text="This was the original 'msg' before 'failed_when' was evaluated.",
                    #     ),
                    # ),
                )

                self.exception = _messages.ErrorSummary(
                    event=event,
                )
        else:
            result = str(value)

            try:
                # The original error's obj should contain the failed expression, which should be origin tagged.
                # This error is intended to provide additional context (the message) for the original error.
                # By repeating the original error's obj here, the error display will group the two errors together.
                raise AnsibleError("A 'failed_when' expression failed.", obj=value.obj) from value
            except Exception as ex:
                wrapped_ex = ex

            self.failed = True
            self.exception = _messages.ErrorSummary(
                # pylint: disable=used-before-assignment
                event=_error_factory.ControllerEventFactory.from_exception(wrapped_ex, _traceback.is_traceback_enabled(_traceback.TracebackEvent.ERROR)),
            )

        self.failed_when_result = result


def _normalize_traceback(value: object | None) -> str | None:
    """Normalize the provided traceback value, returning None if it is falsey."""
    if not value:
        return None

    value = str(value).rstrip()

    if not value:
        return None

    return value + '\n'


@t.final
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class WireTaskResult:
    host_name: str
    task_uuid: str
    task_fields: c.Mapping[str, object]
    utr: UnifiedTaskResult

    @classmethod
    def create(cls, host: Host, task: Task, utr: UnifiedTaskResult) -> t.Self:
        return cls(
            host_name=host.name,
            task_uuid=task._uuid,
            utr=utr,
            task_fields=task.dump_attrs(),
        )


@t.final
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class HostTaskResult:
    host: Host
    task: Task
    utr: UnifiedTaskResult



# Copyright: (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import collections.abc as _c
import functools
import typing as t

from ansible._internal import _collection_proxy
from ansible.module_utils._internal import _messages
from ansible.utils.display import Display

if t.TYPE_CHECKING:
    from ansible._internal import _task
    from ansible.inventory.host import Host
    from ansible.playbook.task import Task


class CallbackTaskResult:
    def __init__(
        self,
        host: Host,
        task: Task,
        utr: _task.UnifiedTaskResult,
    ) -> None:
        self.__host = host
        self.__task = task
        self.__utr = utr

    @property
    def host(self) -> Host:
        """The host associated with this result."""
        return self.__host

    @property
    def _host(self) -> Host:
        """Use the `host` property when supporting only ansible-core 2.19 or later."""
        # deprecated: description='Deprecate `_host` in favor of `host`' core_version='2.23'
        return self.__host

    @property
    def task(self) -> Task:
        """The task associated with this result."""
        return self.__task

    @property
    def _task(self) -> Task:
        """Use the `task` property when supporting only ansible-core 2.19 or later."""
        # deprecated: description='Deprecate `_task` in favor of `task`' core_version='2.23'
        return self.__task

    @property
    def task_fields(self) -> _c.Mapping[str, t.Any]:
        """The task fields associated with this result."""
        Display().deprecated(
            msg="The `CallbackTaskResult.task_fields` mapping is deprecated.",
            help_text="Use `CallbackTaskResult.task` instead.",
            version="2.24",
        )

        return self.__task.dump_attrs()

    @property
    def _task_fields(self) -> _c.Mapping[str, t.Any]:
        """Use the `task_fields` property when supporting only ansible-core 2.19 or later."""
        # deprecated: description='Deprecate `_task_fields` in favor of `task`' core_version='2.23'
        # Display().deprecated(
        #     msg="The `CallbackTaskResult._task_fields` mapping is deprecated.",
        #     help_text="Use `CallbackTaskResult.task` instead.",
        #     version="2.26",
        # )

        return self.__task.dump_attrs()

    @property
    def exception(self) -> _messages.ErrorSummary | None:
        """The error from this task result, if any."""
        return self.__utr.exception

    @property
    def warnings(self) -> _c.Sequence[_messages.WarningSummary]:
        """The warnings for this task, if any."""
        return _collection_proxy.SequenceProxy(self.__utr.warnings or [])

    @property
    def deprecations(self) -> _c.Sequence[_messages.DeprecationSummary]:
        """The deprecation warnings for this task, if any."""
        return _collection_proxy.SequenceProxy(self.__utr.deprecations or [])

    @property
    def task_name(self) -> str:
        return self.task.get_name()

    def is_changed(self) -> bool:
        return self.__utr.changed

    def is_skipped(self) -> bool:
        return bool(self.__utr.skipped)

    def is_failed(self) -> bool:
        return self.__utr.failed

    def is_unreachable(self) -> bool:
        return bool(self.__utr.unreachable)

    @property
    def _result(self) -> _c.MutableMapping[str, t.Any]:
        """Use the `result` property when supporting only ansible-core 2.19 or later."""
        # deprecated: description='Deprecate `_result` in favor of `result`' core_version='2.23'
        return self.result

    @functools.cached_property
    def result(self) -> _c.MutableMapping[str, t.Any]:
        """
        Returns a cached copy of the task result dictionary for consumption by callbacks.
        Internal custom types are transformed to native Python types to facilitate access and serialization.
        """
        # RPFIX-9: FUTURE: consolidate the no_log logic earlier so we don't have to check both the task and UTR here
        return self.__utr.as_result_dict(for_callback=True, censor_callback_result=self.task.no_log or self.__utr.no_log)


TaskResult = CallbackTaskResult
"""Compatibility name for the pre-2.19 callback-shaped TaskResult passed to callbacks."""



from __future__ import annotations

import functools
import typing as t

from .. import _rpc_host
from .._plugins import _strategy

if t.TYPE_CHECKING:
    from .. import _task


def dispatch_to_strategy_result_thread(func: t.Callable) -> t.Callable:
    """Decorator to force a function to be run serially under the strategy result thread."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        request = _rpc_host.AsyncRPCOperation(impl=func, args=args, kwargs=kwargs)

        _strategy.StrategyContext.current().strategy._rpc_queue.put(request)

        request.event.wait()

        return request.result

    return wrapper


class InventoryRPC(_rpc_host.AutoRegisterRPC):
    """
    RPC server implementation for managing inventory from a worker.
    Also used as the type annotation for the RPC client on the worker.
    """

    @staticmethod
    @dispatch_to_strategy_result_thread
    def add_host(host_info: _task.AddHost) -> bool:
        """Add a host to inventory."""
        ctx = _strategy.StrategyContext.current()
        changed = ctx.tqm._inventory.add_dynamic_host(host_info)

        if changed and host_info.host_name not in ctx.strategy._hosts_cache_all:
            ctx.strategy._hosts_cache_all.append(host_info.host_name)

        return changed

    @staticmethod
    @dispatch_to_strategy_result_thread
    def add_group(host_name: str, group_info: _task.AddGroup) -> bool:
        """Add a group to inventory."""
        ctx = _strategy.StrategyContext.current()

        return ctx.tqm._inventory.add_dynamic_group(host_name, group_info)



from __future__ import annotations

from ansible._internal import _task
from ansible._internal._errors._alarm_timeout import AnsibleTimeoutError
from ansible._internal._errors._error_utils import ContributesToTaskResult
from ansible.module_utils.datatag import deprecate_value


class TaskTimeoutError(AnsibleTimeoutError, ContributesToTaskResult):
    """
    A task-specific timeout.

    This exception provides a result dictionary via the ContributesToTaskResult mixin.
    """

    def as_task_result(self, utr: _task.UnifiedTaskResult) -> _task.UnifiedTaskResult:
        help_text = "Configure `DISPLAY_TRACEBACK` to see a traceback on timeout errors."

        frame = deprecate_value(
            value=help_text,
            msg="The `timedout.frame` task result key is deprecated.",
            help_text=help_text,
            version="2.23",
        )

        utr.result_data.update(timedout=dict(frame=frame, period=self.timeout))

        return utr



# -*- coding: utf-8 -*-

# Copyright 2015 Cristian van Ee <cristian at cvee.org>
# Copyright 2015 Igor Gnatenko <i.gnatenko.brain@gmail.com>
# Copyright 2018 Adam Miller <admiller@redhat.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


DOCUMENTATION = """
---
module: dnf
version_added: 1.9
short_description: Manages packages with the I(dnf) package manager
description:
     - Installs, upgrade, removes, and lists packages and groups with the I(dnf) package manager.
options:
  use_backend:
    description:
      - Backend module to use.
    default: "auto"
    choices:
        auto: Automatically select the backend based on the C(ansible_facts.pkg_mgr) fact.
        yum: Alias for V(auto) (see Notes)
        dnf: M(ansible.builtin.dnf)
        yum4: Alias for V(dnf)
        dnf4: Alias for V(dnf)
        dnf5: M(ansible.builtin.dnf5)
    type: str
    version_added: 2.15
  name:
    description:
      - "A package name or package specifier with version, like C(name-1.0).
        When using state=latest, this can be '*' which means run: dnf -y update.
        You can also pass a url or a local path to an rpm file.
        To operate on several packages this can accept a comma separated string of packages or a list of packages."
      - Comparison operators for package version are valid here C(>), C(<), C(>=), C(<=). Example - C(name >= 1.0).
        Spaces around the operator are required.
      - You can also pass an absolute path for a binary which is provided by the package to install.
        See examples for more information.
    aliases:
        - pkg
    type: list
    elements: str
    default: []

  list:
    description:
      - Various (non-idempotent) commands for usage with C(/usr/bin/ansible) and I(not) playbooks.
        Use M(ansible.builtin.package_facts) instead of the O(list) argument as a best practice.
    type: str

  state:
    description:
      - Whether to install (V(present), V(latest)), or remove (V(absent)) a package.
      - Default is V(None), however in effect the default action is V(present) unless the O(autoremove=true),
        then V(absent) is inferred.
    choices: ['absent', 'present', 'installed', 'removed', 'latest']
    type: str

  enablerepo:
    description:
      - C(Repoid) of repositories to enable for the install/update operation.
        These repos will not persist beyond the transaction.
        When specifying multiple repos, separate them with a ",".
    type: list
    elements: str
    default: []

  disablerepo:
    description:
      - C(Repoid) of repositories to disable for the install/update operation.
        These repos will not persist beyond the transaction.
        When specifying multiple repos, separate them with a C(,).
    type: list
    elements: str
    default: []

  conf_file:
    description:
      - The remote dnf configuration file to use for the transaction.
    type: str

  disable_gpg_check:
    description:
      - Whether to disable the GPG checking of signatures of packages being
        installed. Has an effect only if O(state=present) or O(state=latest).
      - This setting affects packages installed from a repository as well as
        "local" packages installed from the filesystem or a URL.
    type: bool
    default: 'no'

  installroot:
    description:
      - Specifies an alternative installroot, relative to which all packages
        will be installed.
    version_added: "2.3"
    default: "/"
    type: str

  releasever:
    description:
      - Specifies an alternative release from which all packages will be
        installed.
    version_added: "2.6"
    type: str

  autoremove:
    description:
      - If V(true), removes all "leaf" packages from the system that were originally
        installed as dependencies of user-installed packages but which are no longer
        required by any such package. Should be used alone or when O(state=absent).
    type: bool
    default: "no"
    version_added: "2.4"
  exclude:
    description:
      - Package name(s) to exclude when O(state=present), or latest. This can be a
        list or a comma separated string.
    version_added: "2.7"
    type: list
    elements: str
    default: []
  skip_broken:
    description:
      - Skip all unavailable packages or packages with broken dependencies
        without raising an error. Equivalent to passing the C(--skip-broken) option.
    type: bool
    default: "no"
    version_added: "2.7"
  update_cache:
    description:
      - Force dnf to check if cache is out of date and redownload if needed.
        Has an effect only if O(state=present) or O(state=latest).
    type: bool
    default: "no"
    aliases: [ expire-cache ]
    version_added: "2.7"
  update_only:
    description:
      - When using latest, only update installed packages. Do not install packages.
      - Has an effect only if O(state=present) or O(state=latest).
    default: "no"
    type: bool
    version_added: "2.7"
  security:
    description:
      - If set to V(true), and O(state=latest) then only installs updates that have been marked security related.
      - Note that, similar to C(dnf upgrade-minimal), this filter applies to dependencies as well.
    type: bool
    default: "no"
    version_added: "2.7"
  bugfix:
    description:
      - If set to V(true), and O(state=latest) then only installs updates that have been marked bugfix related.
      - Note that, similar to C(dnf upgrade-minimal), this filter applies to dependencies as well.
    default: "no"
    type: bool
    version_added: "2.7"
  enable_plugin:
    description:
      - C(Plugin) name to enable for the install/update operation.
        The enabled plugin will not persist beyond the transaction.
    version_added: "2.7"
    type: list
    elements: str
    default: []
  disable_plugin:
    description:
      - C(Plugin) name to disable for the install/update operation.
        The disabled plugins will not persist beyond the transaction.
    version_added: "2.7"
    type: list
    default: []
    elements: str
  disable_excludes:
    description:
      - Disable the excludes defined in DNF config files.
      - If set to V(all), disables all excludes.
      - If set to V(main), disable excludes defined in C([main]) in C(dnf.conf).
      - If set to V(repoid), disable excludes defined for given repo id.
    version_added: "2.7"
    type: str
  validate_certs:
    description:
      - This only applies if using a https url as the source of the rpm. For example, for localinstall.
        If set to V(false), the SSL certificates will not be validated.
      - This should only set to V(false) used on personally controlled sites using self-signed certificates as it avoids verifying the source site.
    type: bool
    default: "yes"
    version_added: "2.7"
  sslverify:
    description:
      - Disables SSL validation of the repository server for this transaction.
      - This should be set to V(false) if one of the configured repositories is using an untrusted or self-signed certificate.
    type: bool
    default: "yes"
    version_added: "2.13"
  allow_downgrade:
    description:
      - Specify if the named package and version is allowed to downgrade
        a maybe already installed higher version of that package.
        Note that setting O(allow_downgrade=true) can make this module
        behave in a non-idempotent way. The task could end up with a set
        of packages that does not match the complete list of specified
        packages to install (because dependencies between the downgraded
        package and others can cause changes to the packages which were
        in the earlier transaction).
      - Since this feature is not provided by C(dnf) itself but by M(ansible.builtin.dnf) module,
        using this in combination with wildcard characters in O(name) may result in an unexpected results.
    type: bool
    default: "no"
    version_added: "2.7"
  download_only:
    description:
      - Only download the packages, do not install them.
    default: "no"
    type: bool
    version_added: "2.7"
  lock_timeout:
    description:
      - Amount of time to wait for the dnf lockfile to be freed.
    required: false
    default: 30
    type: int
    version_added: "2.8"
  install_weak_deps:
    description:
      - Will also install all packages linked by a weak dependency relation.
    type: bool
    default: "yes"
    version_added: "2.8"
  download_dir:
    description:
      - Specifies an alternate directory to store packages.
      - Has an effect only if O(download_only) is specified.
    type: str
    version_added: "2.8"
  allowerasing:
    description:
      - If V(true) it allows erasing of installed packages to resolve dependencies.
    required: false
    type: bool
    default: "no"
    version_added: "2.10"
  nobest:
    description:
      - This is the opposite of the O(best) option kept for backwards compatibility.
      - Since ansible-core 2.17 the default value is set by the operating system distribution.
    required: false
    type: bool
    version_added: "2.11"
  best:
    description:
      - When set to V(true), either use a package with the highest version available or fail.
      - When set to V(false), if the latest version cannot be installed go with the lower version.
      - Default is set by the operating system distribution.
    required: false
    type: bool
    version_added: "2.17"
  cacheonly:
    description:
      - Tells dnf to run entirely from system cache; does not download or update metadata.
    type: bool
    default: "no"
    version_added: "2.12"
extends_documentation_fragment:
- action_common_attributes
- action_common_attributes.flow
attributes:
    action:
        details: dnf has 2 action plugins that use it under the hood, M(ansible.builtin.dnf) and M(ansible.builtin.package).
        support: partial
    async:
        support: none
    bypass_host_loop:
        support: none
    check_mode:
        support: full
    diff_mode:
        support: full
    platform:
        platforms: rhel
notes:
  - When used with a C(loop:) each package will be processed individually, it is much more efficient to pass the list directly to the O(name) option.
  - Group removal doesn't work if the group was installed with Ansible because
    upstream dnf's API doesn't properly mark groups as installed, therefore upon
    removal the module is unable to detect that the group is installed
    U(https://bugzilla.redhat.com/show_bug.cgi?id=1620324).
  - While O(use_backend=yum) and the ability to call the action plugin as
    M(ansible.builtin.yum) are provided for syntax compatibility, the YUM
    backend was removed in ansible-core 2.17 because the required libraries are
    not available for any supported version of Python. If you rely on this
    functionality, use an older version of Ansible.
requirements:
  - python3-dnf
author:
  - Igor Gnatenko (@ignatenkobrain) <i.gnatenko.brain@gmail.com>
  - Cristian van Ee (@DJMuggs) <cristian at cvee.org>
  - Berend De Schouwer (@berenddeschouwer)
  - Adam Miller (@maxamillion) <admiller@redhat.com>
"""

EXAMPLES = """
- name: Install the latest version of Apache
  ansible.builtin.dnf:
    name: httpd
    state: latest

- name: Install Apache >= 2.4
  ansible.builtin.dnf:
    name: httpd >= 2.4
    state: present

- name: Install the latest version of Apache and MariaDB
  ansible.builtin.dnf:
    name:
      - httpd
      - mariadb-server
    state: latest

- name: Remove the Apache package
  ansible.builtin.dnf:
    name: httpd
    state: absent

- name: Install the latest version of Apache from the testing repo
  ansible.builtin.dnf:
    name: httpd
    enablerepo: testing
    state: present

- name: Upgrade all packages
  ansible.builtin.dnf:
    name: "*"
    state: latest

- name: Update the webserver, depending on which is installed on the system. Do not install the other one
  ansible.builtin.dnf:
    name:
      - httpd
      - nginx
    state: latest
    update_only: yes

- name: Install the nginx rpm from a remote repo
  ansible.builtin.dnf:
    name: 'http://nginx.org/packages/centos/6/noarch/RPMS/nginx-release-centos-6-0.el6.ngx.noarch.rpm'
    state: present

- name: Install nginx rpm from a local file
  ansible.builtin.dnf:
    name: /usr/local/src/nginx-release-centos-6-0.el6.ngx.noarch.rpm
    state: present

- name: Install Package based upon the file it provides
  ansible.builtin.dnf:
    name: /usr/bin/cowsay
    state: present

- name: Install the 'Development tools' package group
  ansible.builtin.dnf:
    name: '@Development tools'
    state: present

- name: Autoremove unneeded packages installed as dependencies
  ansible.builtin.dnf:
    autoremove: yes

- name: Uninstall httpd but keep its dependencies
  ansible.builtin.dnf:
    name: httpd
    state: absent
    autoremove: no

- name: Install a modularity appstream with defined stream and profile
  ansible.builtin.dnf:
    name: '@postgresql:9.6/client'
    state: present

- name: Install a modularity appstream with defined stream
  ansible.builtin.dnf:
    name: '@postgresql:9.6'
    state: present

- name: Install a modularity appstream with defined profile
  ansible.builtin.dnf:
    name: '@postgresql/client'
    state: present
"""

import json
import sys

from ansible.module_utils.urls import fetch_file
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.respawn import get_env_with_pythonpath, probe_interpreters_for_module
from ansible.module_utils.embed import EmbedManager
from ansible.module_utils.yumdnf import YumDnf, yumdnf_argument_spec


dnfscript = EmbedManager.embed('..module_utils._embed', 'dnf.py')


class DnfModule(YumDnf):
    """
    DNF Ansible module back-end implementation
    """

    def __init__(self, module):
        # This populates instance vars for all argument spec params
        super(DnfModule, self).__init__(module)

        self._interpreter = self._probe_interpreters()
        self.pkg_mgr_name = "dnf"

    def _build_config(self):
        """Build configuration dictionary for module_utils functions."""
        return {
            'conf_file': self.conf_file,
            'disable_gpg_check': self.disable_gpg_check,
            'sslverify': self.sslverify,
            'installroot': self.installroot,
            'exclude': self.exclude,
            'disable_excludes': self.disable_excludes,
            'releasever': self.releasever,
            'skip_broken': self.skip_broken,
            'nobest': self.nobest,
            'best': self.best,
            'download_only': self.download_only,
            'download_dir': self.download_dir,
            'cacheonly': self.cacheonly,
            'autoremove': self.autoremove,
            'install_weak_deps': self.install_weak_deps,
            'disablerepo': self.disablerepo,
            'enablerepo': self.enablerepo,
            'disable_plugin': self.disable_plugin,
            'enable_plugin': self.enable_plugin,
            'update_cache': self.update_cache,
            'bugfix': self.bugfix,
            'security': self.security,
        }

    def _probe_interpreters(self):
        interpreters = [
            sys.executable,
            '/usr/libexec/platform-python',
            '/usr/bin/python3',
            '/usr/bin/python'
        ]

        # probe well-known system Python locations for accessible bindings, favoring py3
        interpreter = probe_interpreters_for_module(interpreters, module_names=['dnf'])

        if not interpreter:
            # done all we can do, something is just broken (auto-install isn't useful anymore with respawn, so it was removed)
            self.module.fail_json(
                msg='Could not import the dnf python module. '
                    f'Please install `python3-dnf` package. (attempted {interpreters})',
                results=[]
            )

        return interpreter

    def _execute_dnf_script(self, command, config, params=None):
        """Execute the dnf module_utils script via subprocess with JSON RPC."""
        request = {'command': command, 'config': config}
        if params:
            request['params'] = params

        request_json = json.dumps(request)
        python_executable = self._interpreter or sys.executable
        env = get_env_with_pythonpath()

        try:
            rc, stdout, stderr = self.module.run_command(
                [python_executable, '-m', dnfscript.python_module_ref],
                data=request_json,
                check_rc=False,
                handle_exceptions=False,
                environ_update=env,
            )

            if stdout:
                return json.loads(stdout)
            else:
                return {
                    'failed': True,
                    'msg': f'No output from dnf script. stderr: {stderr}',
                    'results': [],
                    'rc': rc
                }
        except json.JSONDecodeError as e:
            return {
                'failed': True,
                'msg': f'Failed to parse JSON from dnf script: {e}. stdout: {stdout}',
                'results': [],
                'rc': 1
            }
        except Exception as e:
            return {
                'failed': True,
                'msg': f'Failed to execute dnf script: {e}',
                'results': [],
                'rc': 1
            }

    def list_items(self, command):
        """List package info based on the command."""
        config = self._build_config()

        result = self._execute_dnf_script(
            command='list',
            config=config,
            params={'list_command': command},
        )

        if result.get('failed'):
            self.module.fail_json(msg=result['msg'], results=[], rc=1)

        for warning in result.get('warnings', []):
            self.module.warn(warning)

        self.module.exit_json(msg='', results=result['results'])

    def ensure(self):
        names = [fetch_file(self.module, name) if '://' in name else name for name in self.names]

        config = self._build_config()
        params = {
            'names': names,
            'state': self.state,
            'autoremove': self.autoremove,
            'update_only': self.update_only,
            'allow_downgrade': self.allow_downgrade,
            'download_only': self.download_only,
            'disable_gpg_check': self.disable_gpg_check,
            'check_mode': self.module.check_mode,
            'download_dir': self.download_dir,
            'allowerasing': self.allowerasing,
        }

        result = self._execute_dnf_script(
            command='ensure',
            config=config,
            params=params,
        )

        for warning in result.get('warnings', []):
            self.module.warn(warning)

        if result.get('failed'):
            error_msg = result.get('msg', 'Unknown error occurred')
            if result.get('failures'):
                self.module.fail_json(
                    msg=error_msg,
                    failures=result['failures'],
                    results=result.get('results', []),
                    rc=result.get('rc', 1)
                )
            else:
                self.module.fail_json(
                    msg=error_msg,
                    results=result.get('results', []),
                    rc=result.get('rc', 1)
                )

        if result.get('changed', False):
            self.module.exit_json(
                msg=result.get('msg', ''),
                changed=True,
                results=result.get('results', []),
                rc=result.get('rc', 0)
            )
        else:
            self.module.exit_json(
                msg=result.get('msg', 'Nothing to do'),
                changed=False,
                results=result.get('results', []),
                rc=result.get('rc', 0)
            )

    def update_cache_only(self):
        config = self._build_config()

        result = self._execute_dnf_script(
            command='update-cache',
            config=config,
        )

        if result.get('failed'):
            self.module.fail_json(
                msg=result['msg'],
                results=[],
                rc=1
            )

        for warning in result.get('warnings', []):
            self.module.warn(warning)

        self.module.exit_json(
            msg='Cache updated',
            changed=result.get('changed', False),
            results=[],
            rc=0
        )

    def run(self):
        if self.update_cache and not self.names and not self.list:
            self.update_cache_only()
            return

        # Set state as installed by default
        # This is not set in AnsibleModule() because the following shouldn't happen
        # - dnf: autoremove=yes state=installed
        if self.state is None:
            self.state = 'installed'

        if self.list:
            self.list_items(self.list)
        else:
            self.ensure()


def main():
    # state=installed name=pkgspec
    # state=removed name=pkgspec
    # state=latest name=pkgspec
    #
    # informational commands:
    #   list=installed
    #   list=updates
    #   list=available
    #   list=repos
    #   list=pkgspec

    yumdnf_argument_spec['argument_spec']['use_backend'] = dict(default='auto', choices=['auto', 'dnf', 'yum', 'yum4', 'dnf4', 'dnf5'])

    module = AnsibleModule(
        **yumdnf_argument_spec
    )

    module_implementation = DnfModule(module)
    module_implementation.run()


if __name__ == '__main__':
    main()



# (c) 2016, Ansible by Red Hat <info@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from ansible.module_utils._internal import _no_six


def pct_to_int(value, num_items, min_value=1):
    """
    Converts a given value to a percentage if specified as "x%",
    otherwise converts the given value to an integer.
    """
    if isinstance(value, str) and value.endswith('%'):
        value_pct = int(value.replace("%", ""))
        return int((value_pct / 100.0) * num_items) or min_value
    else:
        return int(value)


def object_to_dict(obj, exclude=None):
    """
    Converts an object into a dict making the properties into keys, allows excluding certain keys
    """
    if exclude is None or not isinstance(exclude, list):
        exclude = []
    return dict((key, getattr(obj, key)) for key in dir(obj) if not (key.startswith('_') or key in exclude))


def deduplicate_list(original_list):
    """
    Creates a deduplicated list with the order in which each item is first found.
    """
    seen = set()
    return [x for x in original_list if x not in seen and not seen.add(x)]


def __getattr__(importable_name):
    return _no_six.deprecate(importable_name, __name__, "string_types")



"""Galaxy (ansible-galaxy) plugin for integration tests."""

from __future__ import annotations

import os
import tempfile

from ....config import (
    IntegrationConfig,
)

from ....docker_util import (
    docker_cp_from,
    docker_cp_to,
    docker_exec,
)

from ....containers import (
    run_support_container,
)

from ....encoding import (
    to_text,
)

from ....util import (
    display,
)

from . import (
    CloudEnvironment,
    CloudEnvironmentConfig,
    CloudProvider,
)


GALAXY_IMPORTER = b"""
[galaxy-importer]
ansible_local_tmp=~/.ansible/tmp
ansible_test_local_image=false
check_required_tags=false
check_runtime_yaml=false
check_changelog=false
infra_osd=false
local_image_docker=false
log_level_main=INFO
require_v1_or_greater=false
run_ansible_doc=false
run_ansible_lint=false
run_ansible_test=false
run_flake8=false
""".strip()


class GalaxyProvider(CloudProvider):
    """
    Galaxy plugin. Sets up ansible-galaxy servers for tests.
    """

    def __init__(self, args: IntegrationConfig) -> None:
        super().__init__(args)

        self.uses_docker = True
        self.galaxy_image = os.getenv(
            'ANSIBLE_GALAXY_CONTAINER',
            'ghcr.io/ansible/galaxy-ng-test-container:26.03.0'
        )
        self.postgres_image = os.getenv(
            'ANSIBLE_POSTGRES_CONTAINER',
            'public.ecr.aws/docker/library/postgres:13'
        )
        self.amanda_image = os.getenv(
            'ANSIBLE_AMANDA_CONTAINER',
            'ghcr.io/sivel/amanda@sha256:f704fe6f062b8ada59ae6553a70d2175295d068d56f544875980581b7df9c16d'
        )

    def setup(self) -> None:
        """Setup cloud resource before delegation and reg cleanup callback."""
        super().setup()

        # This container is created separately from the actual galaxy container due to
        # needing it created for postgres, but the galaxy container has a dependency on knowing the postgres
        # container id
        #
        # Podman does not extract the image contents until the container starts, this is not true of Docker.
        # The container will start and then immediately exit to conserve resources.
        gdata = run_support_container(
            self.args,
            self.platform,
            self.galaxy_image,
            'galaxy-data',
            [],
            publish_ports=False,
            cmd=['/bin/true'],
            start=True,
            data_container=True,
        )
        if not gdata:
            return

        amanda = run_support_container(
            self.args,
            self.platform,
            self.amanda_image,
            'amanda',
            [8001],
            aliases=['amanda'],
            options=[
                '--volumes-from', gdata.container_id,
            ],
            cmd=['-port', '8001', '-publish'],
        )
        if not amanda:
            return

        postgres = run_support_container(
            self.args,
            self.platform,
            self.postgres_image,
            'galaxy-postgres',
            [5432],
            aliases=['postgres'],
            options=[
                '--volumes-from', gdata.container_id,
            ],
        )
        if not postgres:
            return

        with tempfile.TemporaryDirectory() as tmpdir:
            docker_cp_from(self.args, gdata.container_id, '/galaxy_ng.env', tmpdir)
            galaxy_ng = run_support_container(
                self.args,
                self.platform,
                self.galaxy_image,
                'galaxy_ng',
                [8000, 24816],
                aliases=['galaxy'],
                start=True,
                options=[
                    '--env-file', os.path.join(tmpdir, 'galaxy_ng.env'),
                    '--add-host', f'postgres:{postgres.details.container_ip}',
                ],
                cmd=[
                    '/bin/sh', '-c',
                    '(start-api &); (start-content-app &); start-worker;'
                ],
            )
        if not galaxy_ng:
            return

        injected_files = [
            ('/etc/galaxy-importer/galaxy-importer.cfg', GALAXY_IMPORTER, 'galaxy-importer'),
        ]
        for path, content, friendly_name in injected_files:
            with tempfile.NamedTemporaryFile() as temp_fd:
                temp_fd.write(content)
                temp_fd.flush()
                display.info(f'>>> {friendly_name} Configuration\n{to_text(content)}', verbosity=3)
                docker_exec(self.args, galaxy_ng.container_id, ['mkdir', '-p', os.path.dirname(path)], True, options=['-u', 'root'])
                docker_cp_to(self.args, galaxy_ng.container_id, temp_fd.name, path)
                docker_exec(self.args, galaxy_ng.container_id, ['chown', 'galaxy:galaxy', path], True, options=['-u', 'root'])

        self._set_cloud_config('GALAXY_HOST', 'galaxy')
        self._set_cloud_config('GALAXY_USER', 'admin')
        self._set_cloud_config('GALAXY_PASSWORD', 'admin')
        self._set_cloud_config('AMANDA_HOST', 'amanda')


class GalaxyEnvironment(CloudEnvironment):
    """Galaxy environment plugin. Updates integration test environment after delegation."""

    def get_environment_config(self) -> CloudEnvironmentConfig:
        """Return environment configuration for use in the test environment after delegation."""
        galaxy_user = str(self._get_cloud_config('GALAXY_USER'))
        galaxy_password = str(self._get_cloud_config('GALAXY_PASSWORD'))
        galaxy_host = self._get_cloud_config('GALAXY_HOST')
        amanda_host = self._get_cloud_config('AMANDA_HOST')

        return CloudEnvironmentConfig(
            ansible_vars=dict(
                galaxy_user=galaxy_user,
                galaxy_password=galaxy_password,
                galaxy_ng_server=f'http://{galaxy_host}:8000/api/galaxy',
                amanda=f'http://{amanda_host}:8001',
            ),
            env_vars=dict(
                GALAXY_USER=galaxy_user,
                GALAXY_PASSWORD=galaxy_password,
                GALAXY_NG_SERVER=f'http://{galaxy_host}:8000/api/galaxy',
                AMANDA=f'http://{amanda_host}:8001',
            ),
        )



from __future__ import annotations


class AttributeUnavailableError(Exception):
    """
    This AttributeError-equivalent exception can be raised by custom Jinja-injected objects (e.g. CurrentTask) when an attribute is not yet available.
    It does not extend AttributeError to allow the exception message to be included on the resulting undefined object.
    """



"""Helpers for PowerShell command line and scripting arguments."""

from __future__ import annotations

import base64
import re
import shlex

from . import _clixml

# There are 5 chars that need to be escaped in a single quote.
# https://github.com/PowerShell/PowerShell/blob/b7cb335f03fe2992d0cbd61699de9d9aafa1d7c1/src/System.Management.Automation/engine/parser/CharTraits.cs#L265-L272
_PWSH_SINGLE_QUOTES = re.compile("(['\u2018\u2019\u201a\u201b])")

# These chars are a known set of chars that are safe to use as a bare command.
_PWSH_BARE_CMD = re.compile(r'^[\w\-\\\.\[\]/_~!%^:]+$')


def _get_encoded_arguments(
    script: str,
    args: list[str] | None = None,
) -> list[str]:
    """Builds an -EncodedCommand cmdline argument

    Returns a list of arguments that can be used with PowerShell to invoke
    a script with optional arguments.

    :param script: The script to encode.
    :param args: The string arguments to encode.
    :returns: The encoded arguments string value as a base64 string.
    """
    encoded_cmd = base64.b64encode(script.encode('utf-16-le')).decode()
    pwsh_args = ['-EncodedCommand', encoded_cmd]
    if not args:
        return pwsh_args

    # -EncodedCommand does not accept normal positional parameters. Instead it
    # only supports extra args through an encoded CLIXML string.
    clixml = _clixml.build_array_list_clixml(args)
    encoded_args = base64.b64encode(clixml.encode('utf-16-le')).decode()
    pwsh_args.extend(['-EncodedArguments', encoded_args])

    return pwsh_args


def parse_encoded_cmdline(
    cmd: str,
) -> tuple[str, list[str]] | None:
    """Parses a PowerShell encoded command line string.

    Attempts to parse a command line string for the encoded command/arguments.
    Will return None if not present or if the command line is not a valid for
    PowerShell.

    :param cmd: The command line string to parse.
    :returns: A tuple of the decoded command and list of arguments or None if ill-formed.
    """
    cmd_split = shlex.split(cmd)

    try:
        enc_cmd_idx = cmd_split.index("-EncodedCommand")
    except ValueError:
        enc_cmd_idx = -1

    if enc_cmd_idx == -1 or len(cmd_split) <= enc_cmd_idx + 1:
        return None

    enc_cmd_raw = cmd_split[enc_cmd_idx + 1]
    enc_cmd = base64.b64decode(enc_cmd_raw).decode('utf-16-le')

    try:
        enc_arg_idx = cmd_split.index("-EncodedArguments")
    except ValueError:
        return enc_cmd, []

    if len(cmd_split) <= enc_arg_idx + 1:
        # Was ill-formed so assume it's not a valid encoded pwsh cmd.
        return None

    clixml_raw = cmd_split[enc_arg_idx + 1]
    clixml = base64.b64decode(clixml_raw).decode('utf-16-le')
    cmd_args = _clixml.extract_clixml_strings(clixml)

    return enc_cmd, cmd_args


def get_pwsh_encoded_cmdline(
    script: str,
    *,
    args: list[str] | None = None,
    pwsh_path: str = 'powershell',
    disable_input: bool = False,
    override_execution_policy: bool = False,
) -> list[str]:
    """Builds a PowerShell command line argument list.

    Builds the encoded command line arguments for running the provided script.

    :param script: The PowerShell script to encode.
    :param args: Optional positional arguments to the script.
    :param pwsh_path: The PowerShell executable path, defaults to powershell.
    :param disable_input: Tells PowerShell to not read input from stdin.
    :param override_execution_policy: Adds args to override the execution policy.
    :returns: The command line arguments as a list.
    """
    cmd_args = [pwsh_path, '-NoProfile', '-NonInteractive']
    if override_execution_policy:
        cmd_args.extend(['-ExecutionPolicy', 'Unrestricted'])

    if disable_input:
        cmd_args.extend(['-InputFormat', 'None'])

    encoded_args = _get_encoded_arguments(script, args)
    cmd_args.extend(encoded_args)

    return cmd_args


def build_pwsh_cmd_statement(
    command: str,
    args: list[str] | None = None,
) -> str:
    """Builds a PowerShell command statement.

    Builds a valid PowerShell command statement with optional arguments. This
    statement can be used in a PowerShell script to execute a command with the
    arguments provided.

    :param command: The PowerShell command to execute.
    :param args: The arguments to provide to the command.
    :returns: The PowerShell script as a string representing this command statement.
    """
    pwsh_cmd = quote_pwsh_argument(command)
    if pwsh_cmd != command:
        # If we quoted the cmd we need to add the call operator for pwsh
        # to treat it as a command.
        pwsh_cmd = f"& {pwsh_cmd}"

    if args:
        remaining_args = " ".join([quote_pwsh_argument(a) for a in args])
        pwsh_cmd += f" {remaining_args}"

    return pwsh_cmd


def quote_pwsh_argument(
    value: str,
    *,
    force_quote: bool = False,
) -> str:
    """Quotes a value for use as a PowerShell argument.

    This safely quotes the provided value for use in a PowerShell script. The
    value will either be a bare keyword or a quoted string that is safe to use
    in a PowerShell script.

    :param value: The value to quote.
    :param force_quote: Force the quotes even if the value doesn't need it.
    :returns: The value that is safe to use as a PowerShell argument.
    """
    if not force_quote and _PWSH_BARE_CMD.match(value):
        return value

    # Escaping a pwsh string for single quotes is to just double up on the
    # single quote values inside the string.
    return f"'{_PWSH_SINGLE_QUOTES.sub(r'\1\1', value)}'"



"""
Internal Worker->Controller RPC support module.

Implements a minimal multiprocessing context for an in-process LocalManager RPC server hosted on the primary Ansible controller process.
"""

from __future__ import annotations

import collections.abc as c
import dataclasses
import signal
import threading
import typing as t

from multiprocessing.context import BaseContext
from multiprocessing.managers import BaseManager, Server
from multiprocessing.process import BaseProcess

from ..module_utils._internal._concurrent._futures import DaemonThreadPoolExecutor

_server_ready: threading.Event = threading.Event()


class LocalNotAProcess(BaseProcess):
    """Minimal BaseProcess impl that runs `target` locally in a thread instead of a subprocess."""

    def __init__(self, *posargs, target, args, **kwargs):
        super().__init__(*posargs, **kwargs)

        self._args = args
        self._kwargs = kwargs
        self._target = target
        self._tpe = DaemonThreadPoolExecutor()

    def start(self):
        if threading.current_thread() is not threading.main_thread():
            # temporary signal patch is only safe under the main thread; guaranteed to be restored before it can be used
            raise RuntimeError("Local RPC server must be started from the main thread.")

        original_signal = signal.signal

        try:
            # signal.signal raises if called from a non-main thread, disable it until we're past that point of the server startup
            signal.signal = lambda *args, **kwargs: None

            # the only target this should see is _run_server
            # start cannot return until Server.serve_forever is called (our custom subclass sets the _server_ready event)
            self._tpe.submit(self._target, *self._args, **self._kwargs)

            if not _server_ready.wait(5):
                raise TimeoutError("Local RPC server did not start.")
        finally:
            signal.signal = original_signal  # always restore default signal impl


class LocalContext(BaseContext):
    """Minimal Context impl to support in-proc LocalManager."""

    Process = LocalNotAProcess


class LocalServer(Server):
    """Customized Server for in-proc LocalManager."""

    def serve_forever(self):
        _server_ready.set()
        return super().serve_forever()


class LocalManager(BaseManager):
    """Customized BaseManager for in-proc usage."""

    _Server = LocalServer  # override BaseManager to use custom Server subclass
    _shared_instance: t.ClassVar[t.Self | None] = None
    _shared_lock = threading.Lock()

    def __init__(self):
        type(self)._current = self  # HACK: ew

        super().__init__(serializer='pickle', ctx=LocalContext())

    @property
    def authkey(self) -> bytes:
        return self._authkey

    @classmethod
    def register(
        cls,
        typeid: str,
        callable: t.Callable | None = None,
        proxytype: type | None = None,
        exposed: c.Sequence[str] | None = None,
        method_to_typeid: c.Mapping[str, str] | None = None,
        create_method: bool = True,
    ) -> None:
        with cls._shared_lock:
            super().register(
                typeid=typeid,
                callable=callable,
                proxytype=proxytype,
                exposed=exposed,
                method_to_typeid=method_to_typeid,
                create_method=create_method,
            )

    @classmethod
    def shared_instance(cls) -> t.Self:
        """Access a lazily-created LocalManager singleton."""
        if not cls._shared_instance:
            with cls._shared_lock:
                if not cls._shared_instance:
                    instance = cls()
                    instance.start()

                    cls._shared_instance = instance

        return cls._shared_instance


@dataclasses.dataclass(kw_only=True)
class AsyncRPCOperation:
    """Wrapper around a deferred invocation with a waitable completion event."""

    event: threading.Event = dataclasses.field(default_factory=threading.Event)
    impl: t.Callable
    args: tuple
    kwargs: dict[str, object]

    _response: object = None
    _exception: BaseException | None = None

    # FUTURE: reimplement as a lighter-weight future/asyncio executor if RPC volume gets significantly higher

    @property
    def result(self) -> object:
        if self._exception:
            raise self._exception

        return self._response

    def dispatch(self) -> None:
        try:
            self._response = self.impl(*self.args, **self.kwargs)
        except BaseException as ex:
            self._exception = ex

        self.event.set()


class AutoRegisterRPC:
    """Base class for an RPC implementation which automatically registers its methods."""

    def __init_subclass__(cls, **kwargs):
        instance = cls()
        LocalManager.register(cls.__name__, lambda: instance)

    @classmethod
    def get_client(cls) -> t.Self:
        """Get the RPC client for this implementation."""
        rpc_host = LocalManager.shared_instance()  # this assumes the caller was forked from the controller after the manager was started

        rpc_client = BaseManager(address=rpc_host.address, authkey=rpc_host.authkey)
        rpc_client.register(cls.__name__)
        rpc_client.connect()

        return getattr(rpc_client, cls.__name__)()



from __future__ import annotations

import dataclasses
import typing as t

from ansible._internal._errors import _error_utils
from ansible.errors import AnsibleRuntimeError
from ansible.module_utils._internal import _messages

if t.TYPE_CHECKING:
    from ansible._internal import _task


class AnsibleCapturedError(AnsibleRuntimeError):
    """An exception representing error detail captured in another context where the error detail must be serialized to be preserved."""

    context: str

    def __init__(
        self,
        *,
        obj: t.Any = None,
        event: _messages.Event,
    ) -> None:
        super().__init__(
            obj=obj,
        )

        self._event = event


class AnsibleResultCapturedError(AnsibleCapturedError, _error_utils.ContributesToTaskResult):
    """
    An exception representing error detail captured in a foreign context where an action/module result dictionary is involved.

    This exception provides a result dictionary via the ContributesToTaskResult mixin.
    """

    def __init__(self, event: _messages.Event, utr: _task.UnifiedTaskResult) -> None:
        self.context = utr._captured_error_context
        self._default_message = utr._captured_error_message

        super().__init__(event=event)

        self._utr = utr

    def as_task_result(self, utr: _task.UnifiedTaskResult) -> _task.UnifiedTaskResult:
        return self._utr  # RPFIX-5: DOC: explain this better - Drop the provided UTR on the floor and use the stored one instead.


@dataclasses.dataclass(**_messages._dataclass_kwargs)
class CapturedErrorSummary(_messages.ErrorSummary):
    # RPFIX-5: DOC: clean this up (naming, docstring, etc.)
    error_message: str
    error_context: str
    is_module: bool



"""Helpers for PowerShell's CLIXML data"""

from __future__ import annotations

import base64
import re
import xml.etree.ElementTree as ET

# This is weird, we are matching on byte sequences that match the utf-16-be
# matches for '_x(a-fA-F0-9){4}_'. The \x00 and {4} will match the hex sequence
# when it is encoded as utf-16-be byte sequence.
_STRING_DESERIAL_FIND = re.compile(rb"\x00_\x00x((?:\x00[a-fA-F0-9]){4})\x00_")

# Finds _x in a case insensitive way, _x is the escape sequence for a CLIXML
# str so needs to be escaped first.
_STRING_SERIAL_ESCAPE_ESCAPE = re.compile("(?i)_(x)")

# Finds C0, C1, and surrogate pairs in a unicode string for us to encode
# according to the PSRP rules.
_STRING_SERIAL_ESCAPE = re.compile("[\u0000-\u001f\u007f-\u009f\ud800-\ud8ff\udc00-\udfff\U00010000-\U0010ffff]")

_VT_COLOR_PATTERN = re.compile(rb"\x1b\[[0-9;]*m")


def replace_stderr_clixml(stderr: bytes) -> bytes:
    """Replace CLIXML with stderr data.

    Tries to replace an embedded CLIXML string with the actual stderr data. If
    it fails to parse the CLIXML data, it will return the original data. This
    will replace any line inside the stderr string that contains a valid CLIXML
    sequence.

    :param bytes stderr: The stderr to try and decode.
    :returns: The stderr with the decoded CLIXML data or the original data.
    """
    clixml_header = b"#< CLIXML"

    # Instead of checking both patterns we just see if the next char
    # is \r or \n to match both Windows and POSIX newline after the marker.
    clixml_idx = stderr.find(clixml_header)
    if clixml_idx == -1:
        return stderr

    newline_idx = clixml_idx + 9
    if len(stderr) < (newline_idx + 1) or stderr[newline_idx] not in (ord(b'\r'), ord(b'\n')):
        return stderr

    lines: list[bytes] = []
    is_clixml = False
    for line in stderr.splitlines(True):
        if is_clixml:
            is_clixml = False

            # If the line does not contain the closing CLIXML tag, we just
            # add the found header line and this line without trying to parse.
            end_idx = line.find(b"</Objs>")
            if end_idx == -1:
                lines.append(clixml_header)
                lines.append(line)
                continue

            clixml = line[: end_idx + 7]
            remaining = line[end_idx + 7 :]

            # While we expect the stderr to be UTF-8 encoded, we fallback to
            # the most common "ANSI" codepage used by Windows cp437 if it is
            # not valid UTF-8.
            try:
                clixml_text = clixml.decode("utf-8")
            except UnicodeDecodeError:
                clixml_text = clixml.decode("cp437")

            try:
                decoded_clixml = extract_clixml_strings(clixml_text, stream="Error")
                lines.append("".join(decoded_clixml).encode('utf-8', errors='surrogatepass'))
                if remaining:
                    lines.append(remaining)

            except Exception:
                # Any errors and we just add the original CLIXML header and
                # line back in.
                lines.append(clixml_header)
                lines.append(line)

        elif line.startswith(clixml_header):
            # The next line should contain the full CLIXML data.
            clixml_header = line  # Preserve original newlines value.
            is_clixml = True

        else:
            lines.append(line)

    # This should never happen but if there was a CLIXML header without a newline
    # following it, we need to add it back.
    if is_clixml:
        lines.append(clixml_header)

    # PowerShell 7 is not consistent at all with disabling VT color codes,
    # especially in the CLIXML stderr. Instead of trying to hack it in through
    # env vars that may or may not work we just strip it from the output.
    return _VT_COLOR_PATTERN.sub(b"", b"".join(lines))


def extract_clixml_strings(
    data: str,
    stream: str | None = None,
) -> list[str]:
    """
    Takes a string that contains a CLIXML <Objs> element and extracts any
    string elements within. This is a rudimentary extraction designed for
    stderr CLIXML and -EncodedArguments.

    :param data: The raw CLIXML string.
    :param stream: The optional string to extra the data for.
    :returns: A list of CLIXML strings encoded within the CLIXML string.
    """
    lines: list[str] = []

    # A serialized string will serialize control chars and surrogate pairs as
    # _xDDDD_ values where DDDD is the hex representation of a big endian
    # UTF-16 code unit. As a surrogate pair uses 2 UTF-16 code units, we need
    # to operate our text replacement on the utf-16-be byte encoding of the raw
    # text. This allows us to replace the _xDDDD_ values with the actual byte
    # values and then decode that back to a string from the utf-16-be bytes.
    def rplcr(matchobj: re.Match) -> bytes:
        match_hex = matchobj.group(1)
        hex_string = match_hex.decode("utf-16-be")
        return base64.b16decode(hex_string.upper())

    # There are some scenarios where the stderr contains a nested CLIXML element like
    # '<# CLIXML\r\n<# CLIXML\r\n<Objs>...</Objs><Objs>...</Objs>'.
    # Parse each individual <Objs> element and add the error strings to our stderr list.
    # https://github.com/ansible/ansible/issues/69550
    while data:
        start_idx = data.find("<Objs ")
        end_idx = data.find("</Objs>")
        if start_idx == -1 or end_idx == -1:
            break

        end_idx += 7
        current_element = data[start_idx:end_idx]
        data = data[end_idx:]

        clixml = ET.fromstring(current_element)
        namespace_match = re.match(r'{(.*)}', clixml.tag)
        namespace = f"{{{namespace_match.group(1)}}}" if namespace_match else ""

        entries = clixml.findall(".//%sS" % namespace)
        if not entries:
            continue

        # If this is a new CLIXML element, add a newline to separate the messages.
        if lines:
            lines.append("\r\n")

        for string_entry in entries:
            actual_stream = string_entry.attrib.get('S', None)
            if actual_stream != stream:
                continue

            b_line = (string_entry.text or "").encode("utf-16-be")
            b_escaped = re.sub(_STRING_DESERIAL_FIND, rplcr, b_line)

            lines.append(b_escaped.decode("utf-16-be", errors="surrogatepass"))

    return lines


def build_array_list_clixml(
    values: list[str],
) -> str:
    """Builds a CLIXML string representing a System.Collections.ArrayList of strings."""

    def rplcr(matchobj: object) -> str:
        surrogate_char = matchobj.group(0)
        byte_char = surrogate_char.encode("utf-16-be", errors="surrogatepass")
        hex_char = base64.b16encode(byte_char).decode()
        hex_split = [hex_char[i : i + 4] for i in range(0, len(hex_char), 4)]

        return "".join([f"_x{i}_" for i in hex_split])

    objs = ET.Element('Objs', xmlns="http://schemas.microsoft.com/powershell/2004/04", Version="1.1.0.1")
    obj = ET.SubElement(objs, 'Obj', RefId="0")

    tn = ET.SubElement(obj, 'TN', RefId="0")
    ET.SubElement(tn, 'T').text = "System.Collections.ArrayList"
    ET.SubElement(tn, 'T').text = "System.Object"

    lst = ET.SubElement(obj, 'LST')
    for v in values:
        # Before running the translation we need to make sure that '_x' is
        # escaped as '_x005F_x'. While MS-PSRP doesn't state this, the x is
        # case insensitive so we need to escape both '_x' and '_X'.
        v = re.sub(_STRING_SERIAL_ESCAPE_ESCAPE, r"_x005F_\1", v)

        # Escape any control or codepoints that are represented as a
        # surrogate pair in UTF-16.
        v = re.sub(_STRING_SERIAL_ESCAPE, rplcr, v)
        ET.SubElement(lst, 'S').text = v

    return ET.tostring(objs, encoding='unicode')



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import random
import secrets
import string
import warnings

from dataclasses import dataclass

from ansible import constants as C
from ansible.errors import AnsibleError, AnsibleAssertionError
from ansible.module_utils.common.text.converters import to_text, to_bytes
from ansible.utils.display import Display

PASSLIB_E = None
PASSLIB_AVAILABLE = False

try:
    # deprecated: description='warning suppression only required for Python 3.12 and earlier' python_version='3.12'
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message="'crypt' is deprecated and slated for removal in Python 3.13", category=DeprecationWarning)

        import passlib
        import passlib.hash
        from passlib.utils.handlers import HasRawSalt, PrefixWrapper
        try:
            from passlib.utils.binary import bcrypt64
        except ImportError:
            from passlib.utils import bcrypt64

    PASSLIB_AVAILABLE = True
except Exception as e:
    PASSLIB_E = e

CRYPT_E = None
HAS_CRYPT = False
try:
    from ansible._internal._encryption._crypt import CryptFacade
    _crypt_facade = CryptFacade()
    HAS_CRYPT = True
except Exception as e:
    CRYPT_E = e


display = Display()

__all__ = ['do_encrypt']

DEFAULT_PASSWORD_LENGTH = 20


def random_password(length=DEFAULT_PASSWORD_LENGTH, chars=C.DEFAULT_PASSWORD_CHARS, seed=None):
    """Return a random password string of length containing only chars

    :kwarg length: The number of characters in the new password.  Defaults to 20.
    :kwarg chars: The characters to choose from.  The default is all ascii
        letters, ascii digits, and these symbols ``.,:-_``
    """
    if not isinstance(chars, str):
        raise AnsibleAssertionError(f'{chars=!r} ({type(chars)}) is not a {type(str)}.')

    if seed is None:
        random_generator = secrets.SystemRandom()
    else:
        random_generator = random.Random(seed)

    return u''.join(random_generator.choice(chars) for dummy in range(length))


_SALT_CHARS = string.ascii_letters + string.digits + './'
_VALID_SALT_CHARS = frozenset(_SALT_CHARS)


def random_salt(length=8):
    """Return a text string suitable for use as a salt for the hash functions we use to encrypt passwords.
    """
    # Note passlib salt values must be pure ascii so we can't let the user
    # configure this
    return random_password(length=length, chars=_SALT_CHARS)


@dataclass(frozen=True)
class _Algo:
    crypt_id: str
    salt_size: int
    implicit_rounds: int | None = None
    salt_exact: bool = False
    implicit_ident: str | None = None
    rounds_format: str | None = None
    requires_gensalt: bool = False


class BaseHash(object):
    algorithms = {
        'md5_crypt': _Algo(crypt_id='1', salt_size=8),
        'bcrypt': _Algo(crypt_id='2b', salt_size=22, implicit_rounds=12, salt_exact=True, implicit_ident='2b', rounds_format='cost'),
        'sha256_crypt': _Algo(crypt_id='5', salt_size=16, implicit_rounds=535000, rounds_format='rounds'),
        'sha512_crypt': _Algo(crypt_id='6', salt_size=16, implicit_rounds=656000, rounds_format='rounds'),
    }

    def __init__(self, algorithm):
        self.algorithm = algorithm
        display.vv(f"Using {self.__class__.__name__} to hash input with {algorithm!r}")


class CryptHash(BaseHash):
    algorithms = {
        **BaseHash.algorithms,
        'yescrypt': _Algo(crypt_id='y', salt_size=16, implicit_rounds=5, rounds_format='cost', requires_gensalt=True, salt_exact=True),
    }

    def __init__(self, algorithm: str) -> None:
        super(CryptHash, self).__init__(algorithm)

        if not HAS_CRYPT:
            raise AnsibleError("crypt cannot be used as the 'libxcrypt' library is not installed or is unusable.") from CRYPT_E

        if algorithm not in self.algorithms:
            raise AnsibleError(f"crypt does not support {self.algorithm!r} algorithm")

        self.algo_data = self.algorithms[algorithm]

        if self.algo_data.requires_gensalt and not _crypt_facade.has_crypt_gensalt:
            raise AnsibleError(f"{self.algorithm!r} algorithm requires libxcrypt")

    def hash(self, secret: str, salt: str | None = None, salt_size: int | None = None, rounds: int | None = None, ident: str | None = None) -> str:
        rounds = self._rounds(rounds)
        ident = self._ident(ident)

        if _crypt_facade.has_crypt_gensalt:
            saltstring = self._gensalt(ident, rounds, salt, salt_size)
        else:
            saltstring = self._build_saltstring(ident, rounds, salt, salt_size)

        return self._hash(secret, saltstring)

    def _validate_salt_size(self, salt_size: int | None) -> int:
        if salt_size is not None and not isinstance(salt_size, int):
            raise TypeError('salt_size must be an integer')
        salt_size = salt_size or self.algo_data.salt_size
        if self.algo_data.salt_exact and salt_size != self.algo_data.salt_size:
            raise AnsibleError(f"invalid salt size supplied ({salt_size}), expected {self.algo_data.salt_size}")
        elif not self.algo_data.salt_exact and salt_size > self.algo_data.salt_size:
            raise AnsibleError(f"invalid salt size supplied ({salt_size}), expected at most {self.algo_data.salt_size}")
        return salt_size

    def _salt(self, salt: str | None, salt_size: int | None) -> str:
        salt_size = self._validate_salt_size(salt_size)
        ret = salt or random_salt(salt_size)
        if not set(ret).issubset(_VALID_SALT_CHARS):
            raise AnsibleError("invalid characters in salt")
        if self.algo_data.salt_exact and len(ret) != self.algo_data.salt_size:
            raise AnsibleError(f"invalid salt size supplied ({len(ret)}), expected {self.algo_data.salt_size}")
        elif not self.algo_data.salt_exact and len(ret) > self.algo_data.salt_size:
            raise AnsibleError(f"invalid salt size supplied ({len(ret)}), expected at most {self.algo_data.salt_size}")
        return ret

    def _rounds(self, rounds: int | None) -> int | None:
        return rounds or self.algo_data.implicit_rounds

    def _ident(self, ident: str | None) -> str | None:
        return ident or self.algo_data.crypt_id

    def _gensalt(self, ident: str, rounds: int | None, salt: str | None, salt_size: int | None) -> str:
        if salt is None:
            salt_size = self._validate_salt_size(salt_size)
            rbytes = secrets.token_bytes(salt_size)
        else:
            salt = self._salt(salt, salt_size)
            rbytes = to_bytes(salt)

        prefix = f'${ident}$'
        count = rounds or 0

        try:
            salt_bytes = _crypt_facade.crypt_gensalt(to_bytes(prefix), count, rbytes)
            return to_text(salt_bytes, errors='strict')
        except (NotImplementedError, ValueError) as e:
            raise AnsibleError(f"Failed to generate salt for {self.algorithm!r} algorithm") from e

    def _build_saltstring(self, ident: str, rounds: int | None, salt: str | None, salt_size: int | None) -> str:
        salt = self._salt(salt, salt_size)
        saltstring = f'${ident}' if ident else ''
        if rounds:
            if self.algo_data.rounds_format == 'cost':
                # musl libc requires a 2-digit bcrypt cost ($2b$05$, not $2b$5$).
                saltstring += f'${rounds:02d}'
            else:
                saltstring += f'$rounds={rounds}'
        saltstring += f'${salt}'
        return saltstring

    def _hash(self, secret: str, saltstring: str) -> str:
        try:
            result = _crypt_facade.crypt(to_bytes(secret), to_bytes(saltstring))
        except (OSError, ValueError) as e:
            raise AnsibleError(f"crypt does not support {self.algorithm!r} algorithm") from e

        return to_text(result, errors='strict')


class PasslibHash(BaseHash):
    def __init__(self, algorithm):
        super(PasslibHash, self).__init__(algorithm)

        if not PASSLIB_AVAILABLE:
            raise AnsibleError(f"The passlib Python package must be installed to hash with the {algorithm!r} algorithm.") from PASSLIB_E

        try:
            self.crypt_algo = getattr(passlib.hash, algorithm)
        except Exception:
            raise AnsibleError(f"Installed passlib version {passlib.__version__} does not support the {algorithm!r} algorithm.") from None

    def hash(self, secret, salt=None, salt_size=None, rounds=None, ident=None):
        salt = self._clean_salt(salt)
        rounds = self._clean_rounds(rounds)
        ident = self._clean_ident(ident)
        if salt_size is not None and not isinstance(salt_size, int):
            raise TypeError("salt_size must be an integer")
        return self._hash(secret, salt=salt, salt_size=salt_size, rounds=rounds, ident=ident)

    def _clean_ident(self, ident):
        ret = None
        if not ident:
            if self.algorithm in self.algorithms:
                return self.algorithms.get(self.algorithm).implicit_ident
            return ret
        if self.algorithm == 'bcrypt':
            return ident
        return ret

    def _clean_salt(self, salt):
        if not salt:
            return None
        elif issubclass(self.crypt_algo.wrapped if isinstance(self.crypt_algo, PrefixWrapper) else self.crypt_algo, HasRawSalt):
            ret = to_bytes(salt, encoding='ascii', errors='strict')
        else:
            ret = to_text(salt, encoding='ascii', errors='strict')

        # Ensure the salt has the correct padding
        if self.algorithm == 'bcrypt':
            ret = bcrypt64.repair_unused(ret)

        return ret

    def _clean_rounds(self, rounds):
        algo_data = self.algorithms.get(self.algorithm)
        if rounds:
            return rounds
        elif algo_data and algo_data.implicit_rounds:
            # The default rounds used by passlib depend on the passlib version.
            # For consistency ensure that passlib behaves the same as crypt in case no rounds were specified.
            # Thus use the crypt defaults.
            return algo_data.implicit_rounds
        else:
            return None

    def _hash(self, secret, salt, salt_size, rounds, ident):
        # Not every hash algorithm supports every parameter.
        # Thus create the settings dict only with set parameters.
        settings = {}
        if salt:
            settings['salt'] = salt
        if salt_size:
            settings['salt_size'] = salt_size
        if rounds:
            settings['rounds'] = rounds
        if ident:
            settings['ident'] = ident

        # starting with passlib 1.7 'using' and 'hash' should be used instead of 'encrypt'
        try:
            if hasattr(self.crypt_algo, 'hash'):
                result = self.crypt_algo.using(**settings).hash(secret)
            elif hasattr(self.crypt_algo, 'encrypt'):
                result = self.crypt_algo.encrypt(secret, **settings)
            else:
                raise ValueError(f"Installed passlib version {passlib.__version__} is not supported.")
        except ValueError as ex:
            raise AnsibleError("Could not hash the secret.") from ex

        # passlib.hash should always return something or raise an exception.
        # Still ensure that there is always a result.
        # Otherwise an empty password might be assumed by some modules, like the user module.
        if not result:
            raise AnsibleError(f"Failed to hash with passlib using the {self.algorithm!r} algorithm.")

        # Hashes from passlib.hash should be represented as ascii strings of hex
        # digits so this should not traceback.  If it's not representable as such
        # we need to traceback and then block such algorithms because it may
        # impact calling code.
        return to_text(result, errors='strict')


def do_encrypt(result, algorithm, salt_size=None, salt=None, ident=None, rounds=None):
    if HAS_CRYPT and algorithm in CryptHash.algorithms:
        return CryptHash(algorithm).hash(result, salt=salt, salt_size=salt_size, rounds=rounds, ident=ident)
    elif PASSLIB_AVAILABLE:
        # TODO: deprecate passlib
        return PasslibHash(algorithm).hash(result, salt=salt, salt_size=salt_size, rounds=rounds, ident=ident)
    elif not PASSLIB_AVAILABLE and algorithm not in CryptHash.algorithms:
        # When passlib support is removed, this branch can be removed too
        raise AnsibleError(f"crypt does not support {algorithm!r} algorithm")
    raise AnsibleError("Unable to encrypt nor hash, either libxcrypt (recommended), crypt, or passlib must be installed.") from CRYPT_E



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 Matt Martz <matt@sivel.net>
# Copyright (C) 2015 Rackspace US, Inc.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import re


REJECTLIST_DIRS = frozenset(('.git', 'test', '.github', '.idea'))
SYS_EXIT_REGEX = re.compile(r'[^#]*sys.exit\s*\(.*')
NO_LOG_REGEX = re.compile(r'(?:pass(?!ive)|secret|token|key)', re.I)

# Everything that should not be used in a dictionary of a return value,
# since it will make user's life harder.
FORBIDDEN_DICTIONARY_KEYS = frozenset([
    'clear',
    'copy',
    'fromkeys',
    'get',
    'items',
    'keys',
    'pop',
    'popitem',
    'setdefault',
    'update',
    'values',
])


REJECTLIST_IMPORTS = {
    'requests': {
        'new_only': True,
        'error': {
            'code': 'use-module-utils-urls',
            'msg': ('requests import found, should use '
                    'ansible.module_utils.urls instead')
        }
    },
    r'boto(?:\.|$)': {
        'new_only': True,
        'error': {
            'code': 'use-boto3',
            'msg': 'boto import found, new modules should use boto3'
        }
    },
}


PLUGINS_WITH_RETURN_VALUES = ('module', )
PLUGINS_WITH_EXAMPLES = ('module', )
PLUGINS_WITH_YAML_EXAMPLES = ('module', )



# -*- coding: utf-8 -*-
# Copyright 2015 Cristian van Ee <cristian at cvee.org>
# Copyright 2015 Igor Gnatenko <i.gnatenko.brain@gmail.com>
# Copyright 2018 Adam Miller <admiller@redhat.com>
# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# NOTE: This file intentionally supports a wider variety of python versions than does
# ansible.module_utils.basic, specifically on the lower bounds. The intent is for this
# to work against the system python where the dnf python bindings are supported.
#
# Requires-Python: ">=3.6"

import json
import os
import sys

try:
    import dnf
    import dnf.const
    import dnf.exceptions
    import dnf.module.module_base
    import dnf.package
    import dnf.subject
    import dnf.util

    HAS_DNF = True
except ImportError:
    HAS_DNF = False


class _DnfScriptError(Exception):
    """Exception raised for DNF script errors with structured error information."""

    def __init__(self, msg, failures=None, results=None, rc=1):
        """Initialize DNF script error."""
        self.msg = msg
        self.failures = failures or []
        self.results = results or []
        self.rc = rc
        super().__init__(msg)

    def to_dict(self):
        """Convert exception to dict format for JSON response."""
        return {'changed': False, 'msg': self.msg, 'failures': self.failures, 'results': self.results, 'rc': self.rc, 'failed': True}


def _package_to_envra(package):
    """Convert a dnf.package.Package object to ENVRA string (epoch always included)."""
    return f'{package.epoch}:{package.name}-{package.version}-{package.release}.{package.arch}'


class _DnfJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder that handles dnf-specific objects."""

    def default(self, o):
        if isinstance(o, dnf.package.Package):
            return self._package_to_dict(o)
        return super().default(o)

    @staticmethod
    def _package_to_dict(package):
        """Convert a dnf.package.Package object to a dictionary."""
        result = {
            'name': package.name,
            'arch': package.arch,
            'epoch': f'{package.epoch}',
            'release': package.release,
            'version': package.version,
            'repo': package.repoid,
        }

        result['envra'] = _package_to_envra(package)
        result['nevra'] = result['envra']

        if package.installtime == 0:
            result['yumstate'] = 'available'
        else:
            result['yumstate'] = 'installed'

        return result


def _exit_json(response):
    """Output JSON response and exit with appropriate code."""
    json.dump(response, sys.stdout, cls=_DnfJSONEncoder)
    sys.exit(1 if response.get('failed') else 0)  # pylint: disable=ansible-bad-function


def _configure_base(base, config_dict):
    """Configure the dnf Base object. Returns list of warnings."""
    warnings = []
    conf = base.conf

    conf_file = config_dict.get('conf_file')
    if conf_file:
        if not os.access(conf_file, os.R_OK):
            raise _DnfScriptError(f'cannot read configuration file: {conf_file}')
        conf.config_file_path = conf_file

    conf.read()
    conf.debuglevel = 0

    disable_gpg_check = config_dict.get('disable_gpg_check', False)
    conf.gpgcheck = not disable_gpg_check
    conf.localpkg_gpgcheck = not disable_gpg_check
    conf.assumeyes = True

    sslverify = config_dict.get('sslverify', True)
    conf.sslverify = sslverify

    installroot = config_dict.get('installroot', '/')
    if not os.path.isdir(installroot):
        raise _DnfScriptError(f'Installroot {installroot} must be a directory')

    conf.installroot = installroot
    conf.substitutions.update_from_etc(installroot)

    exclude = config_dict.get('exclude', [])
    if exclude:
        _excludes = list(conf.exclude)
        _excludes.extend(exclude)
        conf.exclude = _excludes

    disable_excludes = config_dict.get('disable_excludes')
    if disable_excludes:
        _disable_excludes = list(conf.disable_excludes)
        if disable_excludes not in _disable_excludes:
            _disable_excludes.append(disable_excludes)
            conf.disable_excludes = _disable_excludes

    releasever = config_dict.get('releasever')
    if releasever is not None:
        conf.substitutions['releasever'] = releasever

    if conf.substitutions.get('releasever') is None:
        warnings.append('Unable to detect release version (use "releasever" option to specify release version)')
        conf.substitutions['releasever'] = ''

    for opt in ('cachedir', 'logdir', 'persistdir'):
        conf.prepend_installroot(opt)

    skip_broken = config_dict.get('skip_broken', False)
    if skip_broken:
        conf.strict = 0

    nobest = config_dict.get('nobest')
    best = config_dict.get('best')
    if nobest is not None:
        conf.best = not nobest
    elif best is not None:
        conf.best = best

    download_only = config_dict.get('download_only', False)
    if download_only:
        conf.downloadonly = True
        download_dir = config_dict.get('download_dir')
        if download_dir:
            conf.destdir = download_dir

    cacheonly = config_dict.get('cacheonly', False)
    if cacheonly:
        conf.cacheonly = True

    autoremove = config_dict.get('autoremove', False)
    conf.clean_requirements_on_remove = autoremove

    install_weak_deps = config_dict.get('install_weak_deps', True)
    conf.install_weak_deps = install_weak_deps

    return warnings


def _configure_repos(base, disablerepo_list, enablerepo_list, disable_gpg_check):
    """Enable and disable repositories matching the provided patterns."""
    repos = base.repos

    for repo_pattern in disablerepo_list:
        if repo_pattern:
            for repo in repos.get_matching(repo_pattern):
                repo.disable()

    for repo_pattern in enablerepo_list:
        if repo_pattern:
            for repo in repos.get_matching(repo_pattern):
                repo.enable()

    if disable_gpg_check:
        for repo in base.repos.iter_enabled():
            repo.gpgcheck = False
            repo.repo_gpgcheck = False


def _init_plugins(base, disable_plugin_list, enable_plugin_list):
    """Initialize and configure plugins."""
    base.setup_loggers()
    base.init_plugins(set(disable_plugin_list), set(enable_plugin_list))
    base.pre_configure_plugins()


def _add_security_filters(base, bugfix, security):
    """Add security and bugfix filters for package upgrades."""
    add_security_filters_method = getattr(base, 'add_security_filters', None)
    if callable(add_security_filters_method):
        filters = {}
        if bugfix:
            filters.setdefault('types', []).append('bugfix')
        if security:
            filters.setdefault('types', []).append('security')
        if filters:
            add_security_filters_method('eq', **filters)
    else:
        # Fallback for older dnf versions
        filters = []
        if bugfix:
            key = {'advisory_type__eq': 'bugfix'}
            filters.append(base.sack.query().upgrades().filter(**key))
        if security:
            key = {'advisory_type__eq': 'security'}
            filters.append(base.sack.query().upgrades().filter(**key))
        if filters:
            base._update_security_filters = filters

    return True


def _is_package_installed(base, package_spec):
    """Check if a package is installed."""
    return bool(dnf.subject.Subject(package_spec).get_best_query(sack=base.sack).installed())


def _is_newer_version_installed(base, pkg_spec):
    """Check if a newer version of the package is already installed."""
    try:
        if isinstance(pkg_spec, dnf.package.Package):
            installed = sorted(base.sack.query().installed().filter(name=pkg_spec.name, arch=pkg_spec.arch))[-1]
            return installed.evr_gt(pkg_spec)
        else:
            solution = dnf.subject.Subject(pkg_spec).get_best_solution(base.sack)
            q = solution['query']
            nevra = solution['nevra']
            if not q or not nevra or nevra.has_just_name() or not nevra.version:
                return False

            # Filter by name and arch (if specified), but NOT by version
            # since we need to find installed packages to compare versions against
            filter_kwargs = {'name': nevra.name}
            if nevra.arch:
                filter_kwargs['arch'] = nevra.arch
            installed = base.sack.query().installed().filter(**filter_kwargs)
            if not installed:
                return False
            return installed[0].evr_gt(q[0])
    except IndexError:
        return False


def _get_modules(module_base, module_spec):
    """Get module information for a module specification."""
    module_spec = module_spec.strip()
    module_list, nsv = module_base._get_modules(module_spec)
    return {
        'module_list': module_list,
        'nsv': {'name': nsv.name if nsv else None, 'stream': nsv.stream if nsv else None, 'version': nsv.version if nsv and hasattr(nsv, 'version') else None},
    }


def _parse_spec_group_file(base, module_base, names, update_only, with_modules):
    """Parse the names list into package specs, group specs, module specs, and filenames."""
    pkg_specs, grp_specs, module_specs, filenames = [], [], [], []
    already_loaded_comps = False

    for name in names:
        if '://' in name:
            filenames.append(name)
        elif name.endswith('.rpm'):
            filenames.append(name)
        elif name.startswith('/'):
            installed = base.sack.query().filter(provides=name, file=name).installed().run()
            if installed:
                pkg_specs.append(installed[0].name)
            elif not update_only:
                pkg_specs.append(name)
        elif name.startswith('@') or ('/' in name):
            if not already_loaded_comps:
                base.read_comps()
                already_loaded_comps = True

            grp_env_mdl_candidate = name[1:].strip()

            if with_modules and module_base:
                mdl_info = _get_modules(module_base, grp_env_mdl_candidate)
                if mdl_info['module_list'] and mdl_info['module_list'][0]:
                    module_specs.append(grp_env_mdl_candidate)
                else:
                    grp_specs.append(grp_env_mdl_candidate)
            else:
                grp_specs.append(grp_env_mdl_candidate)
        else:
            pkg_specs.append(name)

    return pkg_specs, grp_specs, module_specs, filenames


def _update_only_helper(base, pkgs):
    """Handle update_only logic for packages. Returns list of not installed packages."""
    not_installed = []
    for pkg in pkgs:
        if isinstance(pkg, dnf.package.Package):
            pkg_nevra = _package_to_envra(pkg)
        else:
            pkg_nevra = pkg

        if _is_package_installed(base, pkg_nevra):
            try:
                if isinstance(pkg, dnf.package.Package):
                    base.package_upgrade(pkg)
                else:
                    base.upgrade(pkg)
            except Exception as e:
                raise _DnfScriptError(f'Error occurred attempting update_only operation: {e}')
        else:
            not_installed.append(pkg)

    return not_installed


def _install_remote_rpms_helper(base, filenames, update_only, allow_downgrade):
    """Handle installation of remote RPM files."""
    try:
        pkgs = base.add_remote_rpms(filenames)
        if update_only:
            _update_only_helper(base, pkgs)
        else:
            for pkg in pkgs:
                if not (_is_newer_version_installed(base, pkg) and not allow_downgrade):
                    base.package_install(pkg, strict=base.conf.strict)
    except _DnfScriptError:
        raise
    except Exception as e:
        raise _DnfScriptError(f'Error occurred attempting remote rpm operation: {e}')


def _is_module_installed(base, module_base, module_spec, with_modules):
    """Check if a module is installed."""
    if not with_modules:
        return False

    module_info = _get_modules(module_base, module_spec)
    enabled_streams = base._moduleContainer.getEnabledStream(module_info['nsv']['name'])

    if enabled_streams:
        if module_info['nsv']['stream']:
            if module_info['nsv']['stream'] in enabled_streams:
                return True
            else:
                return False
        else:
            return True

    return False


def _mark_package_install(base, pkg_spec, upgrade, allow_downgrade):
    """Mark a package for installation."""
    msg = ''
    strict = base.conf.strict

    try:
        if dnf.util.is_glob_pattern(pkg_spec):
            if upgrade:
                try:
                    base.upgrade(pkg_spec)
                except dnf.exceptions.PackagesNotInstalledError:
                    pass

            try:
                base.install(pkg_spec, strict=strict)
            except dnf.exceptions.MarkingError as e:
                msg = f'No package {pkg_spec} available.'
                if strict:
                    return {'failed': True, 'msg': msg, 'failure': f'{pkg_spec} {e}', 'rc': 1}
            except dnf.exceptions.DepsolveError as e:
                return {'failed': True, 'msg': f'Depsolve Error occurred for package {pkg_spec}.', 'failure': f'{pkg_spec} {e}', 'rc': 1}
            except dnf.exceptions.Error as e:
                return {'failed': True, 'msg': f'Unknown Error occurred for package {pkg_spec}.', 'failure': f'{pkg_spec} {e}', 'rc': 1}
        elif _is_newer_version_installed(base, pkg_spec):
            if allow_downgrade:
                try:
                    base.install(pkg_spec, strict=strict)
                except dnf.exceptions.MarkingError as e:
                    msg = f'No package {pkg_spec} available.'
                    if strict:
                        return {'failed': True, 'msg': msg, 'failure': f'{pkg_spec} {e}', 'rc': 1}
        elif _is_package_installed(base, pkg_spec):
            if upgrade:
                try:
                    base.upgrade(pkg_spec)
                except dnf.exceptions.PackagesNotInstalledError:
                    pass
        else:
            try:
                base.install(pkg_spec, strict=strict)
            except dnf.exceptions.MarkingError as e:
                msg = f'No package {pkg_spec} available.'
                if strict:
                    return {'failed': True, 'msg': msg, 'failure': f'{pkg_spec} {e}', 'rc': 1}
            except dnf.exceptions.DepsolveError as e:
                return {'failed': True, 'msg': f'Depsolve Error occurred for package {pkg_spec}.', 'failure': f'{pkg_spec} {e}', 'rc': 1}
            except dnf.exceptions.Error as e:
                return {'failed': True, 'msg': f'Unknown Error occurred for package {pkg_spec}.', 'failure': f'{pkg_spec} {e}', 'rc': 1}
    except Exception as e:
        return {'failed': True, 'msg': f'Unknown Error occurred for package {pkg_spec}.', 'failure': f'{pkg_spec} {e}', 'rc': 1}

    return {'failed': False, 'msg': msg, 'failure': '', 'rc': 0}


def _sanitize_install_error(spec, error):
    """Sanitize DNF error messages for install operations."""
    if 'no package matched' in str(error) or 'No match for argument:' in str(error):
        return f'No package {spec} available.'
    return error


def _ensure_impl(base, module_base, params):
    """Core implementation of ensure logic."""
    response = {'msg': '', 'changed': False, 'results': [], 'rc': 0}

    failures = []

    names = params.get('names', [])
    state = params.get('state')
    autoremove = params.get('autoremove', False)
    update_only = params.get('update_only', False)
    allow_downgrade = params.get('allow_downgrade', False)
    download_only = params.get('download_only', False)
    disable_gpg_check = params.get('disable_gpg_check', False)
    check_mode = params.get('check_mode', False)
    download_dir = params.get('download_dir')
    allowerasing = params.get('allowerasing', False)
    with_modules = dnf.base.WITH_MODULES and module_base is not None

    if not names and autoremove:
        names = []
        state = 'absent'

    if names == ['*'] and state == 'latest':
        try:
            base.upgrade_all()
        except dnf.exceptions.DepsolveError as e:
            raise _DnfScriptError(msg=f'Depsolve Error occurred attempting to upgrade all packages: {e}', rc=1)
    else:
        pkg_specs, group_specs, module_specs, filenames = _parse_spec_group_file(base, module_base, names, update_only, with_modules)

        pkg_specs = [p.strip() for p in pkg_specs]
        filenames = [f.strip() for f in filenames]
        groups = []
        environments = []

        for group_spec in (g.strip() for g in group_specs):
            group = base.comps.group_by_pattern(group_spec)
            if group:
                groups.append(group.id)
            else:
                environment = base.comps.environment_by_pattern(group_spec)
                if environment:
                    environments.append(environment.id)
                else:
                    raise _DnfScriptError(f'No group {group_spec} available.')

        if state in ['installed', 'present']:
            if filenames:
                _install_remote_rpms_helper(base, filenames, update_only, allow_downgrade)
                for filename in filenames:
                    response['results'].append(f'Installed {filename}')

            if module_specs and with_modules:
                for module in module_specs:
                    if not _is_module_installed(base, module_base, module, with_modules):
                        response['results'].append(f'Module {module} installed.')
                    try:
                        module_base.install([module])
                        module_base.enable([module])
                    except dnf.exceptions.MarkingErrors as e:
                        failures.append(f'{module} {e}')

            for group in groups:
                try:
                    count = base.group_install(group, dnf.const.GROUP_PACKAGE_TYPES)
                    if count == 0:
                        response['results'].append(f'Group {group} already installed.')
                    else:
                        response['results'].append(f'Group {group} installed.')
                except dnf.exceptions.DepsolveError:
                    raise _DnfScriptError(msg=f'Depsolve Error occurred attempting to install group: {group}', failures=failures, results=response['results'])
                except dnf.exceptions.Error as e:
                    failures.append(f'{group} {e}')

            for environment in environments:
                try:
                    base.environment_install(environment, dnf.const.GROUP_PACKAGE_TYPES)
                except dnf.exceptions.DepsolveError:
                    raise _DnfScriptError(
                        msg=f'Depsolve Error occurred attempting to install environment: {environment}', failures=failures, results=response['results']
                    )
                except dnf.exceptions.Error as e:
                    failures.append(f'{environment} {e}')

            if module_specs and not with_modules:
                raise _DnfScriptError(f'No group {module_specs[0]} available.')

            if update_only:
                not_installed = _update_only_helper(base, pkg_specs)
                for spec in not_installed:
                    response['results'].append(f'Packages providing {spec} not installed due to update_only specified')
            else:
                for pkg_spec in pkg_specs:
                    install_result = _mark_package_install(base, pkg_spec, False, allow_downgrade)
                    if install_result['failed']:
                        failures.append(_sanitize_install_error(pkg_spec, install_result['failure']))
                    else:
                        if install_result['msg']:
                            response['results'].append(install_result['msg'])

        elif state == 'latest':
            if filenames:
                _install_remote_rpms_helper(base, filenames, update_only, allow_downgrade)
                for filename in filenames:
                    response['results'].append(f'Installed {filename}')

            if module_specs and with_modules:
                for module in module_specs:
                    try:
                        if _is_module_installed(base, module_base, module, with_modules):
                            module_base.upgrade([module])
                            response['results'].append(f'Module {module} upgraded.')
                        else:
                            module_base.install([module])
                            module_base.enable([module])
                            response['results'].append("Module {0} installed.".format(module))
                    except dnf.exceptions.MarkingErrors as e:
                        failures.append(f'{module} {e}')

            for group in groups:
                try:
                    base.group_upgrade(group)
                    response['results'].append(f'Group {group} upgraded.')
                except dnf.exceptions.CompsError:
                    if not update_only:
                        try:
                            count = base.group_install(group, dnf.const.GROUP_PACKAGE_TYPES)
                            if count == 0:
                                response['results'].append(f'Group {group} already installed.')
                            else:
                                response['results'].append(f'Group {group} installed.')
                        except dnf.exceptions.Error as e2:
                            failures.append(f'{group} {e2}')
                except dnf.exceptions.Error as e:
                    failures.append(f'{group} {e}')

            for environment in environments:
                try:
                    base.environment_upgrade(environment)
                except dnf.exceptions.CompsError:
                    try:
                        base.environment_install(environment, dnf.const.GROUP_PACKAGE_TYPES)
                    except dnf.exceptions.DepsolveError as e2:
                        failures.append(f'{environment} {e2}')
                    except dnf.exceptions.Error as e2:
                        failures.append(f'{environment} {e2}')
                except dnf.exceptions.DepsolveError as e:
                    failures.append(f'{environment} {e}')
                except dnf.exceptions.Error as e:
                    failures.append(f'{environment} {e}')

            if update_only:
                not_installed = _update_only_helper(base, pkg_specs)
                for spec in not_installed:
                    response['results'].append(f'Packages providing {spec} not installed due to update_only specified')
            else:
                for pkg_spec in pkg_specs:
                    install_result = _mark_package_install(base, pkg_spec, True, allow_downgrade)
                    if install_result['failed']:
                        failures.append(_sanitize_install_error(pkg_spec, install_result['failure']))
                    else:
                        if install_result['msg']:
                            response['results'].append(install_result['msg'])

        else:
            if filenames:
                raise _DnfScriptError('Cannot remove paths -- please specify package name.')

            if module_specs and with_modules:
                for module in module_specs:
                    if _is_module_installed(base, module_base, module, with_modules):
                        response['results'].append(f'Module {module} removed.')
                    try:
                        module_base.remove([module])
                    except dnf.exceptions.MarkingErrors as e:
                        failures.append(f'{module} {e}')
                    try:
                        module_base.disable([module])
                    except dnf.exceptions.MarkingErrors as e:
                        failures.append(f'{module} {e}')
                    try:
                        module_base.reset([module])
                    except dnf.exceptions.MarkingErrors as e:
                        failures.append(f'{module} {e}')

            for group in groups:
                try:
                    base.group_remove(group)
                except dnf.exceptions.CompsError as e:
                    response['results'].append(f'{group} {e}')

            for environment in environments:
                try:
                    base.environment_remove(environment)
                except dnf.exceptions.CompsError as e:
                    response['results'].append(f'{environment} {e}')

            for pkg_spec in pkg_specs:
                try:
                    base.remove(pkg_spec)
                except dnf.exceptions.MarkingError as e:
                    response['results'].append(f'{e.value}: {pkg_spec}')

            allowerasing = True

            if autoremove:
                base.autoremove()

    try:
        has_changes = base.resolve(allow_erasing=allowerasing)
    except dnf.exceptions.DepsolveError as e:
        raise _DnfScriptError(msg=f'Depsolve Error occurred: {e}', failures=failures, results=response['results'])
    except dnf.exceptions.Error as e:
        raise _DnfScriptError(msg=f'Unknown Error occurred: {e}', failures=failures, results=response['results'])

    if not has_changes:
        if failures:
            raise _DnfScriptError(msg='Failed to install some of the specified packages', failures=failures, results=response['results'])
        response['msg'] = 'Nothing to do'
        return response
    else:
        response['changed'] = True

        install_action = 'Downloaded' if download_only else 'Installed'
        for package in base.transaction.install_set:
            response['results'].append(f'{install_action}: {package}')
        for package in base.transaction.remove_set:
            response['results'].append(f'Removed: {package}')

        if failures:
            raise _DnfScriptError(msg='Failed to install some of the specified packages', failures=failures, results=response['results'])

        if check_mode:
            response['msg'] = 'Check mode: No changes made, but would have if not in check mode'
            return response

        if download_only and download_dir and base.conf.destdir:
            dnf.util.ensure_dir(base.conf.destdir)
            base.repos.all().pkgdir = base.conf.destdir

        try:
            base.download_packages(base.transaction.install_set)
        except dnf.exceptions.DownloadError as e:
            raise _DnfScriptError(msg=f'Failed to download packages: {e}', failures=failures, results=response['results'])

        if not disable_gpg_check:
            for package in base.transaction.install_set:
                gpgres, gpgerr = base._sig_check_pkg(package)
                if gpgres != 0:  # Not validated successfully
                    if gpgres == 1:  # Need to install cert
                        try:
                            base._get_key_for_package(package)
                        except dnf.exceptions.Error as e:
                            raise _DnfScriptError(f'Failed to validate GPG signature for {package}: {e}')
                    else:  # Fatal error
                        raise _DnfScriptError(f'Failed to validate GPG signature for {package}: {gpgerr}')

        if download_only:
            return response
        else:
            tid = base.do_transaction()
            if tid is not None:
                transaction = base.history.old([tid])[0]
                if transaction.return_code:
                    failures.extend(transaction.output())

        if failures:
            raise _DnfScriptError(msg='Failed to install some of the specified packages', failures=failures, results=response['results'])

        return response


def _setup_base(config):
    """Internal helper to set up and configure a dnf.Base object."""
    base = dnf.Base()

    warnings = _configure_base(base, config)

    _init_plugins(base, config.get('disable_plugin', []), config.get('enable_plugin', []))

    base.read_all_repos()
    _configure_repos(base, config.get('disablerepo', []), config.get('enablerepo', []), config.get('disable_gpg_check', False))

    base.configure_plugins()

    try:
        if config.get('update_cache'):
            base.update_cache()

        base.fill_sack(load_system_repo='auto')
    except Exception:
        base.close()
        raise

    if config.get('bugfix') or config.get('security'):
        _add_security_filters(base, config.get('bugfix', False), config.get('security', False))

    module_base = None
    if dnf.base.WITH_MODULES:
        module_base = dnf.module.module_base.ModuleBase(base)

    return base, module_base, warnings


def list_items(config, command):
    """List packages based on command."""
    base = None
    try:
        base, module_base, warnings = _setup_base(config)

        if command == 'updates':
            command = 'upgrades'

        if command == 'installed':
            results = list(base.sack.query().installed())
        elif command == 'upgrades':
            results = list(base.sack.query().upgrades())
        elif command == 'available':
            results = list(base.sack.query().available())
        elif command in ['repos', 'repositories']:
            results = [{'repoid': repo.id, 'state': 'enabled'} for repo in base.repos.iter_enabled()]
        else:
            results = list(dnf.subject.Subject(command).get_best_query(base.sack))

        return {'results': results, 'warnings': warnings}
    except Exception as e:
        return {'results': [], 'warnings': [], 'failed': True, 'msg': f'{e}'}
    finally:
        if base:
            base.close()


def update_cache_only(config):
    """Update the cache only (no other operations)."""
    base = None
    try:
        base, module_base, warnings = _setup_base(config)
        return {'changed': False, 'warnings': warnings}
    except Exception as e:
        return {'changed': False, 'warnings': [], 'failed': True, 'msg': f'{e}'}
    finally:
        if base:
            base.close()


def ensure(config, params):
    """Main operation for installing, removing, or upgrading packages."""
    base = None
    try:
        base, module_base, warnings = _setup_base(config)

        result = _ensure_impl(base, module_base, params)

        result['warnings'] = warnings
        result['failures'] = []

        return result

    except _DnfScriptError as e:
        return e.to_dict()
    except Exception as e:
        return {'changed': False, 'results': [], 'warnings': [], 'failed': True, 'failures': [], 'rc': 1, 'msg': f'{e}'}
    finally:
        if base:
            base.close()


def main():
    """Main entry point for standalone script execution."""

    if not HAS_DNF:
        _exit_json({'failed': True, 'msg': 'python3-dnf not found'})

    try:
        request = json.load(sys.stdin)
    except Exception as e:
        _exit_json({'failed': True, 'msg': f'Failed to read JSON input: {e}'})

    command = request.get('command')
    if not command:
        _exit_json({'failed': True, 'msg': 'No command specified'})

    config = request.get('config', {})
    params = request.get('params', {})

    if command == 'list':
        list_command = params.get('list_command')
        if not list_command:
            _exit_json({'failed': True, 'msg': 'No list_command specified for list operation'})
        result = list_items(config, list_command)

    elif command == 'ensure':
        result = ensure(config, params)

    elif command == 'update-cache':
        result = update_cache_only(config)

    else:
        _exit_json({'failed': True, 'msg': f'Unknown command: {command}'})

    _exit_json(result)


if __name__ == '__main__':
    main()
